import csv
import os.path
import pathlib
import pickle
import re
import string
import sys
import textwrap
import warnings
from contextlib import contextmanager
from typing import Union

if os.name == 'nt':
    _POSIX = False
    import winreg
else:
    _POSIX = True

_VALID_YESNO_ANSWERS = {"yes": True, "y": True, "ye": True,
                        "no": False, "n": False}
_HOST_RE = re.compile(r"(janeway|ag-bluhm-[0-9]+)(?![.\d])", re.IGNORECASE)
_HOST_SUFFIX = '.physik.rwth-aachen.de'


def _host_append_suffix(match: re.Match) -> str:
    return match.group() + _HOST_SUFFIX


@contextmanager
def changed_directory(new_directory: os.PathLike):
    r"""Temporarily change to a new directory.

    Equivalent to::

        pushd new_directory
        ...
        popd

    Examples
    --------
    >>> import tempfile, os
    >>> print(os.getcwd())  # doctest: +SKIP
    Z:\Code\qutil
    >>> with (
    ...         tempfile.TemporaryDirectory() as tempdir,
    ...         changed_directory(tempdir)
    ... ):
    ...     print(os.getcwd())  # doctest: +SKIP
    C:\Users\HANGLE~1\AppData\Local\Temp\tmpwzayb4vq
    >>> print(os.getcwd())  # doctest: +SKIP
    Z:\Code\qutil
    """

    original_directory = os.getcwd()
    try:
        os.chdir(new_directory)
        yield
    finally:
        os.chdir(original_directory)


def show_saved_figure(path):
    """Unpickle and show a pickled matplotlib figure.

    Parameters
    ----------
    path: str | pathlib.Path
        The path where the figure is saved.

    Returns
    -------
    fig: matplotlib.figure.Figure
        The matplotlib figure instance.

    """
    fig = pickle.load(open(path, 'rb'))

    import matplotlib.pyplot as plt
    dummy = plt.figure()
    new_manager = dummy.canvas.manager
    new_manager.canvas.figure = fig
    fig.set_canvas(new_manager.canvas)

    plt.show()
    return fig


def save_figure(fig, path):
    """Pickle a matplotlib figure.

    Parameters
    ----------
    fig: matplotlib.figure.Figure
        The matplotlib figure instance to be saved.
    path: str | pathlib.Path
        The path where the figure is saved.

    """
    pickle.dump(fig, open(path, 'wb'))


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
    It must be "yes" (the default), "no" or None (meaning
    an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return _VALID_YESNO_ANSWERS[default]
        elif choice in _VALID_YESNO_ANSWERS:
            return _VALID_YESNO_ANSWERS[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def query_overwrite(filepath: Union[str, os.PathLike], default: str = 'no') -> pathlib.Path:
    """Query if file at filepath should be overwritten if it exists.

    Parameters
    ----------
    filepath : Union[str, os.PathLike]
        The file in question.
    default : str, optional
        The default answer. The default is 'no'.

    Returns
    -------
    filepath : pathlib.Path
        The original filepath if the answer was yes. The input path
        else.

    Raises
    ------
    FileExistsError
        If the file exists and the user pressed Ctrl-C to abort.

    """
    overwrite = _VALID_YESNO_ANSWERS[default]
    while not overwrite and (filepath := pathlib.Path(filepath)).is_file():
        if not (overwrite := query_yes_no(f'File {filepath} exists. Overwrite?', default)):
            try:
                filepath = input('Enter a new file name or path or press Ctrl-C to abort: ')
            except KeyboardInterrupt:
                raise FileExistsError(filepath) from None
    return filepath


def to_global_path(path: Union[str, os.PathLike], warn: bool = False) -> pathlib.Path:
    r"""A Path that automatically replaces network drive labels and host names.

    This is useful for saving files in a way that they can be read from
    anywhere, independent of whether the host computer is known in the
    domain or has been mounted as a network share.

    .. note::
        Starting with Python 3.12, subclassing :class:`pathlib.Path` is
        possible, so this functionality could be more nicely
        implemented: https://github.com/python/cpython/pull/31691.

    Parameters
    ----------
    path : str | os.PathLike
        The path to resolve.
    warn : bool, default False
        Warn if on a posix system and mount points cannot be resolved.

    Examples
    --------
    Assuming ``\\janeway.physik.rwth-aachen.de\User AG Bluhm\Common`` is
    mounted as ``Y:\``:

    >>> to_global_path(r'Y:\GaAs')  # doctest: +SKIP
    WindowsPath('//janeway.physik.rwth-aachen.de/User AG Bluhm/Common/GaAs')

    If Y:/ is instead not a network but a local drive, nothing is resolved:

    >>> to_global_path(r'C:\Users\Tobias')  # doctest: +SKIP
    WindowsPath('C:/Users/Tobias')

    Host names are also resolved:

    >>> to_global_path(r'\\ag-bluhm-51\local_share')  # doctest: +SKIP
    WindowsPath('//ag-bluhm-51.physik.rwth-aachen.de/local_share/')

    But if the path is already the full host path, nothing happens:

    >>> to_global_path(r'\\janeway.physik.rwth-aachen.de\User AG Bluhm')  # doctest: +SKIP
    WindowsPath('//janeway.physik.rwth-aachen.de/User AG Bluhm/')

    """
    # TODO: implement as Path subclass for Python 3.12
    path = pathlib.Path(path).expanduser().absolute()
    if isinstance(path, pathlib.PosixPath):
        if warn:
            warnings.warn('Cannot resolve mount points on Unix', UserWarning)
        return path.resolve()

    root, *rest = path.parts
    if root == os.path.sep:
        # Somehow pathlib drops one slash if there is no child
        root = 2 * root + rest[0]
        rest = rest[1:]
    root = _HOST_RE.sub(_host_append_suffix, root)
    if root[0] in string.ascii_uppercase and root[1] == ':':
        # Resolve a mapped network share
        root = os.path.realpath(root)
    if path != (newpath := pathlib.Path(root, *rest)):
        # Recursively resolve hosts
        return to_global_path(newpath, warn)
    return path


def check_path_length(path: Union[str, os.PathLike], ask_for_new: bool = True) -> pathlib.Path:
    """Check whether a path is too long for Windows.

    Parameters
    ----------
    path : Union[str, os.PathLike]
        The path to be checked.
    ask_for_new : bool, optional
        If True, the user is asked to input a new path if the original
        path was too long. Else, this function basically just gives a
        more informative error than by default. The default is True.

    Raises
    ------
    PathTooLongError
        If the path is too long and ask_for_new is False or input was
        aborted.

    Returns
    -------
    path : pathlib.Path
        The valid path.

    """
    path = pathlib.Path(path)
    if _POSIX:
        # For all intents and purposes no length limit (4095 chars)
        return path

    # Windows, check if more than 256 chars are allowed
    sub_key = 'SYSTEM\\CurrentControlSet\\Control\\FileSystem'
    hkey = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, sub_key)

    if winreg.QueryValueEx(hkey, 'LongPathsEnabled')[0]:
        return path

    while len(str(path)) > 259:
        try:
            if not ask_for_new:
                raise KeyboardInterrupt
            warnings.warn(
                textwrap.dedent(
                    f"""
                    The path {os.path.sep.join((path.drive, path.parts[1], '...', path.parts[-1]))}
                    is too long. Please specify a new path or press Ctrl-C.

                    Note: On Windows, the maximum length for a path is 260 characters [1]. Thus,
                        excluding the terminating null character, 259 are available. You can
                        increase this limit by setting the following registry value to 1:
                        HKEY_LOCAL_MACHINE\\{sub_key}\\LongPathsEnabled

                    [1]: https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation?tabs=registry
                    """
                ),
                UserWarning
            )
            path = to_global_path(input('\nNew path: '))
        except KeyboardInterrupt:
            raise PathTooLongError(path) from None
    return path


class PathTooLongError(FileNotFoundError):
    """Indicates a Windows path is longer than 260(256) characters."""
    pass


class CsvLogger:
    """Logger that only open file for writing. Use pandas.read_csv to read csv files"""
    def __init__(self, filename: str, fieldnames: list, dialect='excel-tab', append=False):
        self.dialect = csv.get_dialect(dialect)
        self.filename = os.path.abspath(filename)
        self.fieldnames = fieldnames

        if os.path.exists(filename):
            if not append:
                raise FileExistsError

            # validate file has the correct header
            with open(filename, 'r') as file:
                reader = csv.DictReader(file, dialect=dialect)

                if reader.fieldnames != fieldnames:
                    raise RuntimeError("Existing file has differing fieldnames", reader.fieldnames, fieldnames)

        else:
            with open(self.filename, 'x') as file:
                csv.DictWriter(file, fieldnames=self.fieldnames, dialect=self.dialect).writeheader()

    def write(self, *args):
        with open(self.filename, 'a+') as file:
            csv.DictWriter(file, fieldnames=self.fieldnames, dialect=self.dialect).writerow(dict(zip(self.fieldnames, args)))



"""This is an interface definition which is used by typechecking because lazy importing currently does no allow that as
described in https://scientific-python.org/specs/spec-0001."""

__version__ = "typing_interface_version_dummy"

# necessary due to https://peps.python.org/pep-0484/#stub-files

# package imports
from . import electronics as electronics
from . import hardware as hardware
from . import measurement as measurement
from . import plotting as plotting
from . import signal_processing as signal_processing
from . import ui as ui

# module imports
from . import caching as caching
from . import const as const
from . import domains as domains
from . import functools as functools
from . import image as image
from . import io as io
from . import itertools as itertools
from . import linalg as linalg
from . import math as math
from . import matlab as matlab
from . import misc as misc
from . import pandas_tools as pandas_tools
from . import parallel as parallel
from . import qcodes as qcodes
from . import qi as qi
from . import random as random
from . import typecheck as typecheck

# use this code to update the stub
def _gen_stub(root=None):
    import pathlib
    if root is None:
        root = pathlib.Path(__file__).parent
    sub_modules = []
    for module in root.glob('*.py'):
        module_name = module.stem
        if module_name.startswith('_'):
            continue
        sub_modules.append(module_name)

    sub_packages = []
    for package_init in root.glob("*/__init__.py"):
        package_dir = package_init.parent
        if package_dir == root:
            continue
        package_name = package_dir.name
        sub_packages.append(package_name)

    module_imports = '\n'.join(f'from . import {module_name} as {module_name}' for module_name in sub_modules)

    package_imports = '\n'.join(f'from . import {package_name} as {package_name}' for package_name in sub_packages)

    return '# package imports\n' + package_imports + '\n\n# module imports\n' + module_imports



"""This module contains signal processing functions that work on data in
real space. For consistency, functions in this module should have the
data to be processed as their first argument while the processed data
should be the sole return argument, i.e.::

    def fun(data, *args, **kwargs) -> processed_data:
        ...

"""
import inspect
import warnings
from typing import Optional, Sequence, Callable, Union, Tuple, Literal, TypeVar

import numpy as np
from scipy import signal, fft

from qutil import math
from qutil.functools import chain, partial, wraps
from qutil.signal_processing import fourier_space
from qutil.typecheck import check_literals

# TODO: replace unsupported imports
try:
    from scipy.signal.spectral import _median_bias, _triage_segments
except ImportError:
    from scipy.signal._spectral_py import _median_bias, _triage_segments

try:
    import numba
except ImportError:
    numba = None

_T = TypeVar('_T')


def _standardize(function):
    """Adds variadic kwargs."""
    try:
        parameters = inspect.signature(function).parameters
    except ValueError:
        # ufunc, https://numpy.org/doc/stable/reference/ufuncs.html#ufuncs-kwargs
        parameters = {'out', 'where', 'axes', 'axis', 'keepdims', 'casting', 'order', 'dtype',
                      'subok', 'signature', 'extobj'}

    @wraps(function)
    def wrapper(x, *args, **kwargs):
        # Filter kwargs that function actually accepts
        kwargs = {k: v for k, v in kwargs.items() if k in parameters}
        return function(x, *args, **kwargs)

    return wrapper


def Id(x: _T, *_, **__) -> _T:
    """The identity mapping."""
    return x


def rms(x, /, out=None, *, axis: Optional[int] = None, where=True, dtype=None, keepdims=False,
        **_) -> np.ndarray:
    """Compute the RMS (root-mean-square).

    See :class:`numpy.ufunc` and the `NumPy reference`_ for
    documentation of the arguments.

    .. _NumPy reference: https://numpy.org/doc/stable/reference/ufuncs.html/

    Examples
    --------
    >>> t = np.linspace(0, 2*np.pi, 1001)
    >>> x = 2*np.sqrt(2)*np.sin(t)
    >>> r = rms(x)
    >>> print(r)  # doctest: +ELLIPSIS
    1.9990007493755...
    >>> np.allclose(r, np.sqrt(x.mean()**2 + x.var()))
    True
    """
    x = np.asanyarray(x)
    N = np.take(x.shape, axis or range(x.ndim)).prod()

    result = math.abs2(x, where=where, dtype=dtype)
    result = np.sum(result, axis=axis, dtype=dtype, out=out, keepdims=keepdims, where=where)
    result = np.sqrt(result, out=out, dtype=dtype)
    result /= np.sqrt(N)
    return result


if numba is not None:
    # nb.guvectorize generates gufuncs with all the kwargs, so only work and result array required.
    def _rms(x, res):
        res[0] = 0
        if np.iscomplexobj(x):
            for i in range(x.shape[0]):
                xx = x[i]
                real = xx.real
                imag = xx.imag
                res += real * real + imag * imag
        else:
            for i in range(x.shape[0]):
                real = x[i].real
                res += real * real
        res[0] = np.sqrt(res[0] / x.shape[0])

    _rms.__doc__ = rms.__doc__
    # Expose both the generated ufunc and the wrapped version that complies with the signature
    # convention.
    rms_ufunc = numba.guvectorize([(numba.float32[:], numba.float32[:]),
                                   (numba.float64[:], numba.float64[:]),
                                   (numba.complex64[:], numba.float32[:]),
                                   (numba.complex128[:], numba.float64[:])],
                                  '(n)->()',
                                  target='parallel',
                                  cache=True)(_rms)
    try:
        rms = _standardize(rms_ufunc)
    except AssertionError:
        # Could not parse rms_ufunc signature, leave rms alone
        pass


@check_literals
def butter_filter(x, fs: float, order: int = 5,
                  btype: Literal['lowpass', 'highpass', 'bandpass'] = 'bandpass', f_min: float = 0,
                  f_max: float = np.inf, **_) -> np.ndarray:
    """Apply a digital Butter filter to the data.

    This function provides a simplified interface to SciPy's `signal`
    functionality by abstracting away some of the API.

    Parameters
    ----------
    x : array_like
        The data to be filtered.
    fs : float
        Sample frequency.
    order : int
        The filter order. Default: 5.
    btype : str
        The filter type. {'bandpass', 'highpass', 'lowpass'}.
    f_min, f_max : float
        The edges for (low-, band-, high-)pass filtering.

    See Also
    --------
    :func:`scipy.signal.butter`
    :func:`scipy.signal.sosfilt`
    """
    if btype == 'lowpass':
        cutoff = f_max
    elif btype == 'highpass':
        cutoff = f_min
    elif btype == 'bandpass':
        cutoff = [f_min, f_max]
    else:
        raise ValueError('Expected btype to be one of {lowpass, highpass, bandpass} but got '
                         f'{btype}')
    return signal.sosfilt(signal.butter(order, cutoff, btype, analog=False, output='sos', fs=fs),
                          x)


def welch(x, fourier_procfn: Optional[Union[Callable, Sequence[Callable]]] = None, fs: float = 1.0,
          window: Union[str, Tuple[np.ndarray, ...]] = 'hann', nperseg: Optional[int] = None,
          noverlap: Optional[int] = None, nfft: Optional[int] = None,
          detrend: Union[str, Callable] = 'constant', normalize: Union[bool, Callable] = False,
          return_onesided: Optional[bool] = None, scaling: str = 'density', axis: int = -1,
          average: str = 'mean', workers: Optional[int] = None,
          **settings) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Use Welch's method to estimate the power spectral density.

    Adapted from :mod:`scipy.signal`, so see that module for parameter
    explanations.

    Unlike the SciPy version, this function allows to perform an
    additional processing step on the Fourier transform of the time data
    ``x``. This is done by applying ``fourier_procfn`` to the FFT'd data
    before the PSD is estimated. ``fourier_procfn`` should be a
    (sequence of) callable(s) with the following signature::

        fourier_procfn(Fx, f, **settings) -> Fxp, fp

    The function defaults to the identity map, so should reproduce the
    SciPy result. If a sequence, functions are applied from left to
    right, i.e., if ``fourier_procfn = [a, b, c]``, then they are
    applied as ``c(b(a(Fx, f, **s), f, **s, f, **s)``.

    By default, the twosided spectrum is computed if x is complex, and
    the transformed data are shifted using :func:`~numpy.fft.fftshift` so
    that they are ordered by monotonously increasing frequencies.

    For undocumented parameters see :func:`scipy.signal.welch`.

    Parameters
    ----------
    fourier_procfn : callable or sequence thereof
        A processing function that acts on the Fourier-transformed
        data (see above).
    normalize : callable or bool, default: False
        Similar to `detrend`, this can be used to normalize each Welch
        segment with some function. If True, the data is normalized by
        its standard deviation (corresponding to the RMS by default as
        the normalization is performed after detrending). This can be
        useful if one wants to compare spectra qualitatively.
    workers : int, optional
        The workers parameter of :func:`scipy:scipy.fft.fft`.

    Returns
    -------
    PSD : ndarray
        The power spectral density.
    f : ndarray
        The discrete FFT frequencies.
    ifft : ndarray
        The timeseries data after processing in Fourier space.

    Examples
    --------
    Same as :func:`scipy.signal.welch` for no fourier_procfn:

    >>> from scipy import signal
    >>> rng = np.random.default_rng()
    >>> x = rng.standard_normal((10, 3, 500))
    >>> np.allclose(signal.welch(x)[1], welch(x)[0])  # returned arguments switched around
    True

    Spectrum of differentiated signal:

    >>> def derivative(x, f, **_):
    ...     return x*2*np.pi*f, f
    >>> S, f, dxdt = welch(x, fourier_procfn=derivative)

    Compare to spectrum of numerical derivative:

    >>> import matplotlib.pyplot as plt
    >>> Sn, fn, dxdtn = welch(np.gradient(x, axis=-1))
    >>> lines = plt.loglog(f, S.mean((0, 1)), fn, Sn.mean((0, 1)))
    """
    if np.iterable(fourier_procfn):
        fourier_procfn = chain(*fourier_procfn, n_args=2)
    else:
        fourier_procfn = chain(fourier_procfn or fourier_space.Id, n_args=2)

    # Default to twosided if x is complex
    return_onesided = return_onesided or not np.iscomplexobj(x)

    axis = int(axis)
    # Ensure we have np.arrays, get outdtype
    # outdtype cannot be complex since we only calculate autocorrelations here.
    x = np.asarray(x)
    outdtype = np.result_type(x.real, np.float32)

    if x.size == 0:
        return (np.empty(x.shape, dtype=outdtype),
                np.empty(x.shape, dtype=outdtype),
                np.empty(x.shape, dtype=outdtype))

    if x.ndim > 1 and axis != -1:
        x = np.moveaxis(x, axis, -1)

    if nperseg is not None:  # if specified by user
        nperseg = int(nperseg)
        if nperseg < 1:
            raise ValueError('nperseg must be a positive integer')

    # parse window; if array like, then set nperseg = win.shape
    win, nperseg = _triage_segments(window, nperseg, input_length=x.shape[-1])

    if nfft is None:
        nfft = nperseg
    elif nfft < nperseg:
        raise ValueError('nfft must be greater than or equal to nperseg.')
    else:
        nfft = int(nfft)

    if noverlap is None:
        noverlap = nperseg//2
    else:
        noverlap = int(noverlap)
    if noverlap >= nperseg:
        raise ValueError('noverlap must be less than nperseg.')

    # Handle detrending and window functions
    if not detrend:
        def detrend_func(d):
            return d
    elif not callable(detrend):
        def detrend_func(d):
            return signal.detrend(d, type=detrend, axis=-1)
    elif axis != -1:
        # Wrap this function so that it receives a shape that it could
        # reasonably expect to receive.
        def detrend_func(d):
            d = np.moveaxis(d, -1, axis)
            d = detrend(d)
            return np.moveaxis(d, axis, -1)
    else:
        detrend_func = detrend

    if not normalize:
        def normalize_func(d):
            return d
    elif not callable(normalize):
        def normalize_func(d):
            # RMS normalization. Assumes detrend has already removed a constant trend.
            return d / rms(d, axis=-1, keepdims=True)
    elif axis != -1:
        def normalize_func(d):
            d = np.moveaxis(d, -1, axis)
            d = normalize(d)
            return np.moveaxis(d, axis, -1)
    else:
        normalize_func = normalize

    if np.result_type(win, np.float32) != outdtype:
        win = win.astype(outdtype)

    if scaling == 'density':
        scale = 1.0 / (fs * (win*win).sum())
    elif scaling == 'spectrum':
        scale = 1.0 / win.sum()**2
    else:
        raise ValueError('Unknown scaling: %r' % scaling)

    if return_onesided:
        if np.iscomplexobj(x):
            sides = 'twosided'
            warnings.warn('Input data is complex, switching to '
                          'return_onesided=False')
        else:
            sides = 'onesided'
    else:
        sides = 'twosided'

    if sides == 'twosided':
        freqs_func = chain(partial(fft.fftfreq, d=1/fs), partial(fft.fftshift, axes=-1))
        fft_func = chain(_fft_helper, partial(fft.fftshift, axes=-1), inspect_kwargs=True)
        ifft_func = chain(partial(fft.ifftshift, axes=-1), fft.ifft)
    else:
        # sides == 'onesided'
        freqs_func = partial(fft.rfftfreq, d=1/fs)
        fft_func = _fft_helper
        ifft_func = fft.irfft

    freqs = freqs_func(nfft)

    # Perform the windowed FFTs. Need to pass kwargs so that FunctionChain can pass on only those
    # that are allowed for each function
    result = fft_func(x, win=win, detrend_func=detrend_func, normalize_func=normalize_func,
                      nperseg=nperseg, noverlap=noverlap, nfft=nfft, sides=sides, workers=workers)

    # Do custom stuff with the Fourier transformed data
    result, freqs = fourier_procfn(result, freqs, **settings)

    # Absolute value square and scaling to get the PSD
    result = scale * math.abs2(result)

    if sides == 'onesided':
        if nfft % 2:
            result[..., 1:] *= 2
        else:
            # Last point is unpaired Nyquist freq point, don't double
            result[..., 1:-1] *= 2

    # Inverse fft for processed time series data (not averaged over Welch segments)
    if not all(fp is fourier_space.Id for fp in fourier_procfn.functions):
        y, _ = fourier_procfn(
            # Basically just fft.fft or fft.rfft
            fft_func(
                x, win=1, detrend_func=Id, normalize_func=Id, nperseg=x.shape[-1], noverlap=0,
                nfft=x.shape[-1], sides=sides, workers=workers
            )[..., 0, :],
            freqs_func(x.shape[-1]),
            **settings
        )
        ifft = np.moveaxis(ifft_func(y, workers=workers), -1, axis)
    else:
        ifft = np.moveaxis(x, -1, axis)

    result = result.astype(outdtype)

    # Output is going to have new last axis for time/window index, so a
    # negative axis index shifts down one
    if axis < 0:
        axis -= 1

    # Roll frequency axis back to axis where the data came from
    result = np.moveaxis(result, -1, axis)

    # Average over windows.
    if len(result.shape) >= 2 and result.size > 0:
        if result.shape[-1] > 1:
            if average == 'median':
                # np.median must be passed real arrays for the desired result
                bias = _median_bias(result.shape[-1])
                if np.iscomplexobj(result):
                    result = (np.median(np.real(result), axis=-1)
                              + 1j * np.median(np.imag(result), axis=-1))
                else:
                    result = np.median(result, axis=-1)
                result /= bias
            elif average == 'mean':
                result = result.mean(axis=-1)
            else:
                raise ValueError('average must be "median" or "mean", got %s'
                                 % (average,))
        else:
            result = np.reshape(result, result.shape[:-1])

    return result, freqs, ifft


def _fft_helper(x, win, detrend_func, normalize_func, nperseg, noverlap, nfft, sides, workers):
    """
    Calculate windowed FFT, for internal use by
    `scipy.signal._spectral_helper`.

    This is a helper function that does the main FFT calculation for
    `_spectral helper`. All input validation is performed there, and the
    data axis is assumed to be the last axis of x. It is not designed to
    be called externally. The windows are not averaged over; the result
    from each window is returned.

    Returns
    -------
    result : ndarray
        Array of FFT data

    Notes
    -----
    Adapted from matplotlib.mlab
    Then adapted from scipy.signal

    .. versionadded:: 0.16.0
    """
    # Created sliding window view of array
    if nperseg == 1 and noverlap == 0:
        result = x[..., np.newaxis]
    else:
        # https://stackoverflow.com/a/5568169
        step = nperseg - noverlap
        result = np.lib.stride_tricks.sliding_window_view(
            x, window_shape=nperseg, axis=-1, writeable=True
        )
        result = result[..., 0::step, :]

    # Detrend and normalize each data segment individually
    result = normalize_func(detrend_func(result))

    # Apply window by multiplication. No inplace op here since result might have uncastable type
    result = win*result

    # Perform the fft. Acts on last axis by default. Zero-pads automatically
    if sides == 'twosided':
        func = fft.fft
    else:
        result = result.real
        func = fft.rfft

    # Can overwrite because a new array is created above
    return func(result, n=nfft, workers=workers, overwrite_x=True)



"""This module contains some useful plotting functions.

TODO
----
    - Use matplotlib widgets for LineClicker, CoordClicker. See
      https://matplotlib.org/stable/gallery/index.html#widgets
"""
from __future__ import annotations

import dataclasses
import os
import pathlib
import warnings
from contextlib import contextmanager
from typing import Sequence, Tuple, Mapping, Any, Union, List, Callable, Literal, TypeVar
from weakref import WeakValueDictionary

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from cycler import cycler
from matplotlib import ticker, colors, scale

from ..typecheck import check_literals

try:
    import qcodes
except ImportError:
    qcodes = None

_color_names = ['blue', 'magenta', 'green', 'orange', 'teal', 'maygreen',
                'red', 'purple', 'violet', 'bordeaux', 'petrol', 'yellow']
_color_tuples = [
    (0/255, 84/255, 159/255),       # blue
    (227/255, 0/255, 102/255),      # magenta
    (87/255, 171/255, 39/255),      # green
    (246/255, 168/255, 0/255),      # orange
    (0/255, 152/255, 161/255),      # teal
    (189/255, 205/255, 0/255),      # maygreen
    (204/255, 7/255, 30/255),       # red
    (122/255, 111/255, 172/255),    # purple
    (97/255, 33/255, 88/255),       # violet
    (161/255, 16/255, 53/255),      # bordeaux
    (0/255, 97/255, 101/255),       # petrol
    (255/255, 237/255, 0/255),      # yellow
]
RWTH_COLORS = dict(zip(_color_names, _color_tuples))

ScaleT = TypeVar('ScaleT', bound=scale.ScaleBase)
NormT = TypeVar('NormT', bound=colors.Normalize)


class CoordClicker:
    def __init__(self, fig, plot_click: bool = False, print_coords: bool = False, precision=3,
                 **style):
        self.fig = fig
        self.print_coords = print_coords
        self.plot_click = plot_click
        self.style = {**{'marker': 'x', 'color': 'r'}, **style}
        self.precision = precision
        self.xs = []
        self.ys = []
        self.cid = self.fig.canvas.mpl_connect('button_press_event', self)

    def __call__(self, click):
        self.xs.append(click.xdata)
        self.ys.append(click.ydata)
        if self.plot_click:
            click.inaxes.plot(self.xs[-1], self.ys[-1], **self.style)
            self.fig.canvas.draw()
        if self.print_coords:
            with np.printoptions(precision=self.precision):
                print(f"(x, y) = {np.array([self.xs, self.ys])[:, -1]}")

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        self.fig.canvas.mpl_disconnect(self.cid)


class LineClicker:
    def __init__(self, ax, print_coords: bool = False, color='tab:red'):
        self.ax = ax
        self.fig = ax.get_figure()
        self.line, = self.ax.plot([], [], color=color)
        self.print_coords = print_coords
        self.xs = list(self.line.get_xdata())
        self.ys = list(self.line.get_ydata())
        self.cid = self.fig.canvas.mpl_connect('button_press_event', self)

    def __call__(self, click):
        if click.inaxes != self.ax:
            return
        self.xs.append(click.xdata)
        self.ys.append(click.ydata)
        self.line.set_data(self.xs, self.ys)
        self.fig.canvas.draw()
        if self.print_coords:
            with np.printoptions(precision=3):
                print(f"(x, y) = {np.array(self.line.get_data())[:, -1]}")

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        self.fig.canvas.mpl_disconnect(self.cid)


class BlitManager:
    """A manager for blitting raster graphics to speed up plot
    refreshes.

    This class is taken from the matplotlib documentation [1]_. See also
    there for a general introduction to blitting and how it can help
    speed up plot updates.

    Parameters
    ----------
    canvas : FigureCanvasAgg
        The canvas to work with, this only works for subclasses of the
        Agg canvas which have the
        :meth:`~FigureCanvasAgg.copy_from_bbox` and
        :meth:`~FigureCanvasAgg.restore_region` methods.

    animated_artists : Iterable[Artist]
        List of the artists to manage

    Examples
    --------
    Also adapted from the documentation:

    >>> import time

    >>> # make a new figure
    >>> fig, ax = plt.subplots()
    >>> # add a line
    >>> x = np.linspace(-10, 10, 1001)
    >>> ln, = ax.plot(x, np.sin(x), animated=True)
    >>> # add a frame number
    >>> fr_number = ax.annotate(
    ...     "frame: 0",
    ...     (0, 1),
    ...     xycoords="axes fraction",
    ...     xytext=(10, -10),
    ...     textcoords="offset points",
    ...     ha="left",
    ...     va="top",
    ...     animated=True,
    ... )
    >>> bm = BlitManager(fig.canvas, [ln, fr_number])
    >>> # make sure our window is on the screen and drawn
    >>> plt.show(block=False)
    >>> plt.pause(.1)

    >>> elapsed = np.empty(201)
    >>> elapsed[0] = time.perf_counter()
    >>> for j in range(200):
    ...     # update the artists
    ...     ln.set_ydata(np.sin(x + (j / 200) * np.pi))
    ...     fr_number.set_text("frame: {j}".format(j=j))
    ...     # tell the blitting manager to do its thing
    ...     bm.update()
    ...     elapsed[j+1] = time.perf_counter()
    >>> print(f'Average time per update: {np.diff(elapsed).mean():.2g} s')  #doctest: +SKIP
    Average time per update: 0.0026 s

    Just for comparison, without blitting:

    >>> ln.set_animated(False)
    >>> ln.set_ydata(np.sin(x))
    >>> fr_number.set_animated(False)
    >>> fr_number.set_text("frame: 0")

    >>> elapsed = np.empty(201)
    >>> elapsed[0] = time.perf_counter()
    >>> for j in range(200):
    ...     # update the artists
    ...     ln.set_ydata(np.sin(x + (j / 200) * np.pi))
    ...     fr_number.set_text("frame: {j}".format(j=j))
    ...     fig.canvas.draw_idle()
    ...     fig.canvas.flush_events()
    ...     elapsed[j+1] = time.perf_counter()
    >>> print(f'Average time per update: {np.diff(elapsed).mean():.2g} s')  #doctest: +SKIP
    Average time per update: 0.015 s

    References
    ----------
    .. [1] https://matplotlib.org/stable/tutorials/advanced/blitting.html
    """

    def __init__(self, canvas, animated_artists=()):
        self.canvas = canvas
        self._bg = None
        self._artists = []

        for a in animated_artists:
            self.add_artist(a)
        # grab the background on every draw
        self.cid = canvas.mpl_connect("draw_event", self.on_draw)

    def on_draw(self, event):
        """Callback to register with 'draw_event'."""
        cv = self.canvas
        if event is not None:
            if event.canvas != cv:
                raise RuntimeError
        self._bg = cv.copy_from_bbox(cv.figure.bbox)
        self._draw_animated()

    def add_artist(self, art):
        """Add an artist to be managed.

        Parameters
        ----------
        art : Artist
            The artist to be added. Will be set to 'animated' (just
            to be safe). *art* must be in the figure associated with
            the canvas this class is managing.

        """
        if art.figure.canvas.figure != self.canvas.figure:
            raise RuntimeError
        art.set_animated(True)
        self._artists.append(art)

    def _draw_animated(self):
        """Draw all of the animated artists."""
        fig = self.canvas.figure
        for a in self._artists:
            fig.draw_artist(a)

    def update(self):
        """Update the screen with animated artists."""
        cv = self.canvas
        fig = cv.figure
        # paranoia in case we missed the draw event,
        if self._bg is None:
            self.on_draw(None)
        else:
            # restore the background
            cv.restore_region(self._bg)
            # draw all of the animated artists
            self._draw_animated()
            # update the GUI state
            cv.blit(fig.bbox)
        # let the GUI event loop process anything it has to do
        cv.flush_events()


def plot_blitted(ax,
                 x: Union[Callable[[], npt.ArrayLike], npt.ArrayLike],
                 y: Union[Callable[[], npt.ArrayLike], npt.ArrayLike],
                 *animated_artists: mpl.artist.Artist,
                 **kwargs) -> Callable[[], None]:
    """Wraps :meth:`~matplotlib:matplotlib.axes.Axes.plot` to use
    :class:`BlitManager`.

    Parameters
    ----------
    ax :
        The axes to plot in.
    x :
        Either array_like or a callback to get new x-values.
    y :
        Either array_like or a callback to get new y-values.
    *animated_artists :
        Additional animated artists for the BlitManager to handle.
    **kwargs :
        Forwarded to :meth:`~matplotlib:matplotlib.axes.Axes.plot`.

    Returns
    -------
    update_callback :
        A callback that updates the plot with new data.

    Examples
    --------
    >>> import time
    >>> def x_callback():
    ...     return [np.cos((time.time() - t0) % 2*np.pi)]
    >>> def y_callback():
    ...     return [np.sin((time.time() - t0) % 2*np.pi)]
    >>> fig, ax = plt.subplots()
    >>> ax.set_aspect('equal')
    >>> ax.grid(True)
    >>> _ = ax.plot(np.cos(np.linspace(0, 2*np.pi, 201)[1:]),
    ...             np.sin(np.linspace(0, 2*np.pi, 201)[1:]),
    ...             color='k')
    >>> txt = "Angle: {:.1f}º"
    >>> angle = ax.annotate(
    ...     txt,
    ...     (0, 1),
    ...     xycoords="axes fraction",
    ...     xytext=(10, -10),
    ...     textcoords="offset points",
    ...     ha="left",
    ...     va="top",
    ...     animated=True,
    ... )
    >>> t0 = time.time()
    >>> update_callback = plot_blitted(ax, x_callback, y_callback,
    ...                                angle, marker='o')
    >>> while time.time() - t0 < 4*np.pi:
    ...     angle.set_text(txt.format(np.rad2deg(
    ...         (time.time() - t0) % 2*np.pi
    ...     )))
    ...     update_callback()

    """
    kwargs['animated'] = True
    line, = ax.plot(x() if callable(x) else x,
                    y() if callable(y) else y,
                    **kwargs)

    manager = BlitManager(ax.figure.canvas, [line, *animated_artists])

    plt.show(block=False)
    plt.pause(0.01)

    def update_callback():
        if callable(x):
            line.set_xdata(x())
        if callable(y):
            line.set_ydata(y())
        manager.update()

    return update_callback


def pcolorfast_blitted(ax, *args,
                       colorbar: bool = False,
                       autoscale: bool = False,
                       animated_artists: Sequence[mpl.artist.Artist] = (),
                       **kwargs) -> Callable[[], None]:
    """Wraps :meth:`~matplotlib:matplotlib.axes.Axes.pcolorfast` to use
    :class:`BlitManager`.

    Parameters
    ----------
    ax :
        The axes to plot in.
    *args :
        Either three-tuple of X, Y, and C, or just C, where X and Y are
        array_like and C is a callback to get new image values.
    colorbar :
        Plot a colorbar.

        Note that the bar is not updated inside loops.
    autoscale :
        Update colorbar scales.
    *animated_artists :
        Additional animated artists for the BlitManager to handle.
    **kwargs :
        Forwarded to :meth:`~matplotlib:matplotlib.axes.Axes.pcolorfast`.

    Returns
    -------
    update_callback :
        A callback that updates the plot with new data.

    Examples
    --------
    >>> import time
    >>> def img_callback():
    ...     arr[:] = np.rot90(arr)
    ...     return arr
    >>> fig, ax = plt.subplots()
    >>> ax.set_aspect('equal')
    >>> arr = np.arange(4).reshape(2, 2)
    >>> update_callback = pcolorfast_blitted(ax, [1, 2], [3, 4], img_callback,
    ...                                      colorbar=True)
    >>> for _ in range(50):
    ...     time.sleep(0.2)
    ...     update_callback()

    """
    *XY, C = args
    kwargs['animated'] = True
    img = ax.pcolorfast(*XY, C(), **kwargs)

    if colorbar:
        ax.figure.colorbar(img, ax=ax)

    manager = BlitManager(ax.figure.canvas, [img, *animated_artists])

    plt.show(block=False)
    plt.pause(0.01)

    def update_callback():
        img.set_data(C())
        if autoscale:
            img.autoscale()
        manager.update()

    return update_callback


def _to_corner_coords(index: pd.Index) -> np.ndarray:
    """Helper function to transform N coordinates pointing at centers of bins to N+1 coords pointing to the edges"""
    coords = index.values

    delta = coords[-1] - coords[-2]

    return np.concatenate((coords, [coords[-1] + delta])) - delta / 2


def _data_hash(*args: np.ndarray) -> int:
    return hash(tuple(arg.tobytes() for arg in args))


@contextmanager
def changed_plotting_backend(backend='agg'):
    """ This decorator/context manager wraps a function to temporarily use the agg backend for matplotlib.

    This class wraps functions to use the agg backend for matplotlib plots. It saves the previously chosen backend,
    sets agg as the backend, runs the wrapped function, and then sets the backend back to what it was previously. The
    standard backend leaks memory, thus switching to agg is advantageous if one only want to savefig things.
    """
    original_backend = mpl.get_backend()
    plt.switch_backend(backend)
    try:
        yield
    finally:
        plt.switch_backend(original_backend)


def plot_2d_dataframe(df: pd.DataFrame,
                      ax: plt.Axes = None, square=True,
                      column=None, index_is_y=True,
                      update_mode: str = 'auto',
                      colorbar_kwargs: Mapping[str, Any] = None,
                      pcolormesh_kwargs: Mapping[str, Any] = None,
                      colorbar_powerlimits: Tuple[int, int] = None) -> plt.Axes:
    """Plot pandas data frames using pcolormesh. This function expects numeric labels and it can update an existing
    plot. Have a look at seaborn.heatmap if you need something else.

    'auto': 'rescale' if x-label, y-label, and title are the same else 'clear'
    'clear': Clear axis before drawing
    'overwrite': Just plot new data frame on top (no colorbar is drawn)
    'rescale': Recalculate and redraw the colorbar

    Details:
     - plotted meshes are stored in ax.meshes
     - The colorbar is stored in ax.custom_colorbar (DO NOT RELY ON THIS)
     - If the plotted data is already present we just shift it to the top using set_zorder
     - Uses _data_hash(x, y, c) to identify previously plotted data

    :param df: pandas dataframe to plot
    :param ax: Axes object
    :param square:
    :param column: Select this column from the dataframe and unstack the index
    :param index_is_y: If true the index are on the y-axis and the columns on the x-axis
    :param update_mode: 'auto',  'overwrite' or 'rescale'
    :param colorbar_kwargs: passed to plt.colorbar
    :param pcolormesh_kwargs: passed to plt.pcolormesh
    :param colorbar_powerlimits: passed to `colorbar.formatter.set_powerlimits` if not None. Pass (0, 0) to enforce scientific notation
    :return:
    """
    if ax is None:
        ax = plt.gca()

    if square:
        ax.set_aspect("equal")

    if column is None and len(df.columns) == 1 and len(df.index.levshape) == 2:
        column = df.columns[0]

    if column is not None:
        title = column
        series = df[column]
        df = series.unstack()

    else:
        title = None

    colorbar_kwargs = colorbar_kwargs or {}
    pcolormesh_kwargs = pcolormesh_kwargs or {}

    c = df.values
    x_idx = df.columns
    y_idx = df.index

    if not index_is_y:
        c = np.transpose(c)
        x_idx, y_idx = y_idx, x_idx

    x_label = x_idx.name
    y_label = y_idx.name

    if update_mode == 'auto':
        if (x_label, y_label, title) == (ax.get_xlabel(), ax.get_ylabel(), ax.get_title()):
            update_mode = 'rescale'
        else:
            update_mode = 'clear'
    if update_mode not in ('clear', 'rescale', 'overwrite'):
        raise ValueError('%s is an invalid value for update_mode' % update_mode)

    if update_mode == 'clear':
        if hasattr(ax, 'custom_colorbar'):
            # clear colorbar axis
            ax.custom_colorbar.ax.clear()
        ax.clear()
        ax.meshes = WeakValueDictionary()

    y = _to_corner_coords(y_idx)
    x = _to_corner_coords(x_idx)

    if not hasattr(ax, 'meshes'):
        ax.meshes = WeakValueDictionary()

    df_hash = _data_hash(x, y, c)
    current_mesh = ax.meshes.get(df_hash, None)

    if current_mesh is None:
        # data not yet drawn -> draw it
        current_mesh = ax.pcolormesh(x, y, c, **pcolormesh_kwargs)
        ax.meshes[df_hash] = current_mesh

    # push to foreground
    max_z = max(mesh.get_zorder() for mesh in ax.meshes.values()) if ax.meshes else 0
    current_mesh.set_zorder(max_z + 1)

    if update_mode != 'overwrite':
        all_data = [mesh.get_array()
                    for mesh in ax.meshes.values()]
        vmin = min(map(np.min, all_data))
        vmax = max(map(np.max, all_data))

        if not hasattr(ax, 'custom_colorbar'):
            ax.custom_colorbar = plt.colorbar(ax=ax, mappable=current_mesh, **colorbar_kwargs)
            if colorbar_powerlimits:
                ax.custom_colorbar.formatter.set_powerlimits(colorbar_powerlimits)

        for mesh in ax.meshes.values():
            mesh.set_clim(vmin, vmax)

        try:
            ax.custom_colorbar.set_clim(vmin, vmax)
        except AttributeError:
            ax.custom_colorbar.mappable.set_clim(vmin, vmax)

    else:
        # TODO: fix
        warnings.warn("for update_mode='overwrite' the colorbar code is stupid")

    ax.set(ylabel=y_label, xlabel=x_label, title=title)

    return ax


def update_plot(handle, data):
    """Update a plot.

    Parameters
    ----------
    handle: matplotlib data handle
        The plot object that is updated. For instance, a lines.Line2D or
        image.AxesImage object.
    *data: Sequence
        New data to plot.
            - for line plots: [xdata, ydata]
            - for image plots: imdata (m x n array)

    """
    handle.set_data(data)
    if hasattr(handle, 'colorbar'):
        handle.colorbar.set_array(data)
        handle.colorbar.changed()
        handle.colorbar.autoscale()
        handle.colorbar.draw_all()

    # Rescale
    handle.axes.relim()
    handle.axes.autoscale_view()
    # We need to draw *and* flush
    handle.figure.canvas.draw()
    handle.figure.canvas.flush_events()


def cycle_plots(plot_callback, *args,
                fig: plt.Figure = None, ax: plt.Axes = None, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
    """Call ``plot_callback(fig, ax, curr_pos, *args, **kwargs)`` on each left/right arrow key press.
    Initially curr_pos = 0. The right arrow increases and the left arrow decreases the current position.
    There is no limit so you need to do the wraparound yourself if needed:

    >>> plot_data = [(x1, y1), (x2, y2), ...]  # doctest: +SKIP
    >>> def example_plot_callback(fig, ax, pos):  # doctest: +SKIP
    ...     idx = pos % len(plot_data)
    ...     ax.plot(*plot_data[idx])
    """
    def key_event(e):
        if e.key == "right":
            key_event.curr_pos += 1
        elif e.key == "left":
            key_event.curr_pos -= 1
        else:
            return
        plot_callback(fig, ax, key_event.curr_pos, *args, **kwargs)
        plt.draw_all()

    key_event.curr_pos = 0

    if fig is None:
        if ax is None:
            fig = plt.figure()
        else:
            fig = ax.get_figure()
    if ax is None:
        ax = fig.add_subplot(111)

    if isinstance(ax, np.ndarray):
        assert all(a in fig.axes for a in ax.flat)
    else:
        assert ax in fig.axes, "axes not in figure"

    fig.canvas.mpl_connect('key_press_event', key_event)

    plot_callback(fig, ax, key_event.curr_pos, *args, **kwargs)
    plt.draw_all()

    return fig, ax


def get_rwth_color_cycle(alpha: float = 1, exclude: Sequence[str] = None):
    """Get the default RWTH color cycle with an alpha channel.

    Parameters
    ----------
    alpha: float
        The alpha (transparency) value for the color cycle.
    exclude: sequence of str
        Exclude these colors. See RWTH_COLORS for all available keys.
        Yellow is a good choice to exclude.

    Example
    -------
    >>> from unittest import mock; plt = mock.MagicMock() # do no do in real code ;)
    >>> cycle_100 = get_rwth_color_cycle(1.)
    >>> cycle_50 = get_rwth_color_cycle(.5)
    >>> ax = plt.gca()
    >>> x = np.linspace(0, 2*np.pi, 51)
    >>> for i, (v100, v50) in enumerate(zip(cycle_100, cycle_50)):
    ...     ax.plot(x, np.sin(x) + i / len(cycle_100), lw=2, **v100)  # doctest: +SKIP
    ...     ax.plot(x, -np.sin(x) + i / len(cycle_50), lw=2, **v50)   # doctest: +SKIP
    ...     pass

    See Also
    --------
    https://matplotlib.org/stable/tutorials/intermediate/color_cycle.html

    https://matplotlib.org/cycler/
    """
    if alpha < 0 or alpha > 1:
        raise ValueError('alpha should be in the range [0, 1].')

    exclude = exclude or []
    return cycler(color=[tup + (alpha,) for name, tup in RWTH_COLORS.items()
                         if name not in exclude])


rwth_color_cycle_25 = get_rwth_color_cycle(.25)
rwth_color_cycle_50 = get_rwth_color_cycle(.5)
rwth_color_cycle_75 = get_rwth_color_cycle(.75)
rwth_color_cycle_100 = get_rwth_color_cycle(1)
rwth_color_cycle = rwth_color_cycle_100


def list_styles(include_matplotlib_styles: bool = False) -> List[str]:
    """List available matplotlib styles.

    Parameters
    ----------
    include_matplotlib_styles : bool, default False
        Include the default styles that ship with ``matplotlib`` as well
        as the custom styles in the matplotlib config directory. If
        False, only the custom styles defined in this module are
        listed.

    Returns
    -------
    styles : list[str]
        The styles such that ``plt.style.use(style)`` works for all
        ``style`` in ``styles``.
    """
    custom_styles = []
    root_path = pathlib.Path(__file__).parents[1]
    for file in (root_path / 'plotting').glob('*.mplstyle'):
        file = file.relative_to(root_path.parent)
        style = '.'.join(file.parent.parts + (file.stem,))
        custom_styles.append(style)

    if include_matplotlib_styles:
        return plt.style.available + custom_styles
    else:
        return custom_styles


@check_literals
def reformat_axis(ax_or_cbar: mpl.axes.Axes | mpl.colorbar.Colorbar, data: npt.ArrayLike,
                  unit: str, which: Literal['x', 'y', 'c']) -> str:
    """Scales an axis with SI prefixes.

    .. note::
        This function requires :mod:`qcodes` to be installed.

    Parameters
    ----------
    ax_or_cbar :
        The matplotlib :class:`~matplotlib:matplotlib.axes.Axes` or
        :class:`~matplotlib:matplotlib.colorbar.Colorbar` to reformat.
    data :
        The data to reformat.
    unit :
        The unit of the data.
    which :
        The type of axis. Either x, y, or c for colorbar axis.

    Returns
    -------
    prefix :
        The SI unit prefix for the reformatted data.

    """
    def _get_scale(formatter):
        return formatter.scale if isinstance(formatter, SIFormatter) else 0

    if qcodes is None:
        raise RuntimeError('This function requires qcodes.')

    if np.isnan(data).all():
        # preempt unhandled exception in find_scale_and_prefix
        return ''

    if which == 'x':
        old_scale = _get_scale(ax_or_cbar.xaxis.get_major_formatter())
    elif which == 'y':
        old_scale = _get_scale(ax_or_cbar.yaxis.get_major_formatter())
    else:
        old_scale = _get_scale(ax_or_cbar.formatter)

    prefix, new_scale = qcodes.plotting.axis_labels.find_scale_and_prefix(data, unit)
    if new_scale == old_scale:
        # Nothing to do
        return prefix

    formatter = SIFormatter(new_scale)
    if which == 'x':
        ax_or_cbar.xaxis.set_major_formatter(formatter)
    elif which == 'y':
        ax_or_cbar.yaxis.set_major_formatter(formatter)
    else:
        ax_or_cbar.formatter = formatter
        ax_or_cbar.update_ticks()
    return prefix


def norm_to_scale(norm: NormT | None) -> ScaleT:
    """Covert a colormap normalization to a linear axis scale.

    This is useful to synchronize scales of 1d slice plots through a 2d
    color plot.

    Currently supported are:

    1. :class:`~matplotlib:matplotlib.colors.Normalize` /
       :class:`~matplotlib:matplotlib.scale.LinearScale`
    2. :class:`~matplotlib:matplotlib.colors.LogNorm` /
       :class:`~matplotlib:matplotlib.scale.LogScale`
    3. :class:`~matplotlib:matplotlib.colors.SymLogNorm` /
       :class:`~matplotlib:matplotlib.scale.SymLogScale`
    4. :class:`~matplotlib:matplotlib.colors.AsinhNorm` /
       :class:`~matplotlib:matplotlib.scale.AsinhScale`

    Examples
    --------
    >>> from matplotlib import colors
    >>> rng = np.random.default_rng()
    >>> data = rng.exponential(size=(1024, 256))
    >>> fig, ax = plt.subplots(1, 2, sharey=True, width_ratios=[1, 3])
    >>> img = ax[1].pcolormesh(data, norm=colors.LogNorm())
    >>> cbar = fig.colorbar(img)
    >>> line, = ax[0].plot(data.sum(axis=1), np.arange(1024))
    >>> ax[0].set_xscale(norm_to_scale(img.norm))

    """
    # The first argument (axis) is only there for backwards compatibility in
    # mpl and is unused.
    if isinstance(norm, colors.AsinhNorm):
        return scale.AsinhScale(None, linear_width=norm.linear_width)
    if isinstance(norm, colors.LogNorm):
        return scale.LogScale(None)
    if isinstance(norm, colors.SymLogNorm):
        return scale.SymmetricalLogScale(None, linthresh=norm.linthresh)
    return scale.LinearScale(None)


def assert_interactive_figure(fig):
    """Asserts that ``fig`` is interactive."""
    try:
        if (
                fig.canvas.required_interactive_framework is None
                and not os.environ.get('GITLAB_CI', False)
        ):
            # https://github.com/matplotlib/matplotlib/issues/20281#issuecomment-846057011
            raise RuntimeError('Please enable an interactive backend.')
    except AttributeError:
        # we tried
        pass


class SIFormatter(ticker.FuncFormatter):

    def __init__(self, scale: int):
        self.scale = scale
        super().__init__(lambda value, pos: f"{value * 10 ** (-scale):g}")


@dataclasses.dataclass
class LaTeXPlotHelper:
    A4_SIZE_INCHES = (8.25, 11.75)

    textwidth_in_inches: float = dataclasses.field(default=0.9 * A4_SIZE_INCHES[0])

    def scale_relative_textwidth(self, fig, width: float, height: Union[bool, float] = False):
        if isinstance(fig, int):
            fig = plt.figure(fig)

        old_w, old_h = fig.get_size_inches()
        new_w = self.textwidth_in_inches * width
        if isinstance(height, float):
            new_h = self.textwidth_in_inches * height
        else:
            new_h = old_h * (new_w / old_h) if height else old_h
        fig.set_size_inches(new_w, new_h)



from __future__ import annotations

import importlib
import inspect
import logging
import abc
import dataclasses
import json
import multiprocessing as mp
import os
import queue
import sys
import tempfile
import threading
import time
from collections import deque
from collections.abc import Callable, Iterable, Sequence
from multiprocessing.managers import SyncManager
from pathlib import Path
from types import ModuleType
from typing import Literal, TypeVar, Mapping, Any, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib import animation, colorbar, scale, patches as mpatches
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import TimerBase, KeyEvent
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpecFromSubplotSpec
from matplotlib.widgets import Button

from .core import reformat_axis, norm_to_scale, assert_interactive_figure
from .. import misc, io, itertools, functools

if sys.version_info >= (3, 10):
    from typing import ParamSpec, Protocol
else:
    from typing_extensions import ParamSpec, Protocol

_LOG = logging.getLogger(__name__)

EVENT_LOOP_TIMEOUT = 1e-3
"""The time the event loop is spun while waiting for new data.

1 ms seems like a short enough time for interaction with the figure to
feel responsive, but feel free to reduce this at cost of CPU usage.
"""
MAX_CLOCK_LENGTH = 100
"""The maximum number of most recent event times to track."""

P = ParamSpec('P')
T = TypeVar('T')
ArtistT = TypeVar('ArtistT', bound=Artist)
"""A matplotlib artist type."""
ScaleT = TypeVar('ScaleT', bound=scale.ScaleBase)
"""A matplotlib scale type."""
TimerT = TypeVar('TimerT', bound=TimerBase)
"""A matplotlib timer type."""
StyleT = Union[str, dict[str, Any], Path]
"""A matplotlib style type allowed in
:func:`matplotlib:matplotlib.style.use`."""


class DataSourceProtocol(Protocol[P, T]):
    """Protocol for data-producing thread target."""

    def __call__(self,
                 data_queue: queue.Queue[T] | mp.Queue[T],
                 stop_event: threading.Event | mp.Event,
                 **kwargs: P.kwargs) -> None:
        pass


class LineOperation(Protocol):
    """Reduction operation from 2D to 1D."""
    label: str
    """An axis label."""

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray: ...


class Clock(deque):
    """A simple clock that can be used as a callback to store call times."""

    def __call__(self):
        self.append(time.perf_counter())


class Stopwatch(deque):
    """A simple stopwatch that stores time differences between two
    successive calls."""
    tic: bool
    """Start of a time interval."""

    @functools.wraps(deque.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tic = True

    @property
    def toc(self) -> bool:
        """Stop of a time interval."""
        return not self.tic

    def __call__(self):
        if self.tic:
            self.append(time.perf_counter())
            self.tic = False
        else:
            self[-1] = time.perf_counter() - self[-1]
            self.tic = True


@dataclasses.dataclass
class OneTimeFlag:
    """A single-use flag that resets to False if it evaluated to True."""
    _val: bool = False

    def __bool__(self) -> bool:
        if bool(self._val):
            self.unset()
            return True
        return False

    def set(self):
        """Set the flag to True."""
        self._val = True

    def unset(self):
        """Set the flag to False."""
        self._val = False


# TODO: staticmethod when py38 is dropped
def _queue_draw(rescale: Callable[P, bool]) -> Callable[P, None]:
    """Wrap `rescale` to set the flag requesting a draw."""
    # TODO: Set's and unsets the flag unnecessarily when decorated
    #       methods call each other. This should not trigger
    #       multiple draws (hopefully) as the event loop does not
    #       spin further until the first decorated method returns

    @functools.wraps(rescale)
    def wrapped(self: LiveViewT, *args: P.args, **kwargs: P.kwargs):
        # First unset so that draws are deferred until after rescale() ran
        draw_enqueued = bool(self.animation.draw_enqueued)
        draw_required = rescale(self, *args, **kwargs)
        if draw_required or draw_enqueued:
            self.animation.draw_enqueued.set()

    return wrapped


class FuncAnimation(animation.FuncAnimation):
    """Animation with optional redraw of entire figure.

    Set the :attr:`draw_enqueued` flag using :meth:`OneTimeFlag.set` to
    request a redraw of the entire figure at the next frame update.

    Furthermore, this class modifys matplotlib's implementation to draw the
    first frame immediately after starting instead of waiting for one
    `interval` to pass.

    See :class:`~matplotlib:matplotlib.animation.FuncAnimation` for
    parameters and documentation.
    """

    def __init__(self, fig, func, frames=None, init_func=None, fargs=None,
                 save_count=None, *, cache_frame_data=True, event_source=None, **kwargs):

        self._draw_enqueued = OneTimeFlag()
        self._stopwatch = Stopwatch(maxlen=MAX_CLOCK_LENGTH)
        self._LOG = logging.getLogger(f'{__name__}.{self.__class__.__qualname__}')

        super().__init__(fig, func, frames, init_func, fargs, save_count,
                         cache_frame_data=cache_frame_data, event_source=event_source, **kwargs)

    @property
    def draw_enqueued(self) -> OneTimeFlag:
        """Flag indicating that the full figure should be drawn at the
        next occasion."""
        return self._draw_enqueued

    def _post_draw(self, framedata, blit):
        if self.draw_enqueued:
            # It's not clear to me why, but a full draw() is required here
            # instead of a draw_idle(). The latter will only paint over old
            # artists but not remove them.
            self._stopwatch()
            self._fig.canvas.draw()
            self._stopwatch()
        # If blit is not enabled, super()'s method will trigger a draw_idle()
        # but this should not do anything if draw() was called above.
        super()._post_draw(framedata, blit)

    def _start(self, *args):
        # Since _step() is only added right before starting the animation, and
        # we want it to execute once before resetting the interval, we need to
        # add the callback doing that *after* calling super's _start(). This
        # means a tiny delay between starting and drawing the first frame, but
        # oh well.
        super()._start(*args)
        self.event_source = with_interval_once(self.event_source, temporary_interval_ms=0)

    def _step(self, *args):
        # Sometimes the super method returns still_going when the event
        # source has actually already been deleted. Due to the way animations
        # work, at this point in the code we cannot prevent the error from
        # occuring anymore.
        try:
            return super()._step(*args)
        except AttributeError:
            if self.event_source is None:
                self._LOG.error('_step() was called by timer after _stop() had already been '
                                'called. This caused the following error in matplotlib and is '
                                'probably due to timing inaccuracies.')
            raise

    def stop(self, *args):
        # Just to make it public
        super()._stop(*args)


class LiveViewBase(abc.ABC):
    """Base class for live plotting using :mod:`matplotlib`.

    Subclasses can add animated artists by overriding
    :meth:`_add_animated_artists` and define how the artists are
    updated in :meth:`_update`.

    To start the live view, call :meth:`start`.

    .. note::
        Only artists inside axes are animated. Hence, titles and labels
        will not be updated if blitting is enabled.

    Parameters
    ----------
    data_source :
        A callback that puts new data into a :class:`queue.Queue` until
        a :class:`threading.Event` is set. The callback is run in a
        separate thread and should have the following signature::

            (queue.Queue, threading.Event, *, ...) -> None

        That is, they should implement the :class:`DataSourceProtocol`
        interface.

        The required return value depends on the subclass
        implementation of :meth:`_update`.
    update_interval_ms :
        The interval in milliseconds in which to run the matplotlib
        event loop, i.e., the (maximum) update rate of the plot.
    show_fps :
        Show a text annotation showing current and average frame rate.
    useblit :
        Use blitting to increase performance. Can cause issues with
        when data update rates are low and `autoscale` is enabled. For
        more information on blitting, see the matplotlib documentation
        [1]_.
    blocking_queue :
        Make the queue only accept one item before it is received. This
        results in the data production thread waiting for the plotting
        to catch up if the latter is too slow.
    autoscale :
        Automatically rescale the axes specified. Disabled if None or
        empty string.
    autoscale_interval_ms :
        The interval in milliseconds between automatic rescales. If
        None, rescales every time the plot is updated.
    event_source :
        A timer that triggers an update event. Should not be needed by
        users.
    xlabel :
        The x-axis label of the plot.
    ylabel :
        The y-axis label of the plot.
    units :
        Base SI units of the axes. SI prefixes will automatically be
        added when rescaling.
    xlim :
        Static xlims. If None, the limits are dynamic and can be
        rescaled using the button or the `autoscale` setting.
    ylim :
        Static ylims. If None, the limits are dynamic and can be
        rescaled using the button or the `autoscale` setting.
    xscale :
        Scale of the x-axis.
    yscale :
        Scale of the y-axis.
    style :
        A valid option for :func:`plt.style.use`.
    fig :
        A figure to use for the plot.
    ax :
        An axes to use as the main axes of the plot.
    fig_kw :
        Keyword-arguments passed to the figure constructor.
    log_level :
        If not None, set logging level.
    **data_source_kwargs :
        Keyword arguments forwarded to the data callback.

    Examples
    --------
    For examples, see the implementations of this class in this module
    or in :mod:`mjolnir.plotting`.

    References
    ----------
    .. [1] https://matplotlib.org/stable/tutorials/advanced/blitting.html

    """
    # attributes relevant for subclasses.
    animation: FuncAnimation
    """The animation used to facilitate live plotting."""
    event_sources: dict[str, TimerT | object]
    """Event sources generating events for the animation event loop."""
    axes: dict[str, Axes]
    """The axes hosting artists."""
    animated_artists: dict[str, ArtistT]
    """The artists being animated."""
    static_artists: dict[str, ArtistT | Sequence[ArtistT]]
    """The artists that remain static."""
    buttons: dict[str, Button]
    """Clickable buttons."""
    cids: dict[str, int]
    """Event identifiers to keep track of."""
    thread: threading.Thread
    """The thread pushing data to the queue."""
    _xlabel_axes: Axes
    """Subclasses can set this attribute to specify the axes whose
    xlabel should be set."""
    _ylabel_axes: Axes
    """Subclasses can set this attribute to specify the axes whose
    ylabel should be set."""
    _event_clock: Clock
    """Tracks the times at which frames were requested."""

    @dataclasses.dataclass
    class LiveViewProxy:
        """A proxy class returned by :meth:`LiveViewBase.in_process`
        handling communication with the child process."""

        process: mp.Process
        data_thread: threading.Thread
        data_queue: mp.Queue
        stop_event: mp.Event

        def stop(self):
            """Stop the data stream from the main process without killing
            the view object."""
            self.stop_event.set()
            while not self.data_queue.empty():
                self.data_queue.get(block=False)
            try:
                self.data_thread.join(timeout=1)
            except RuntimeError:
                # Thread not started
                pass

        def attach(self, data_source: DataSourceProtocol, **data_source_kwargs):
            """Attach a new data source to the view and start it in a
            thread of the main process.

            The old thread will be terminated if still running.
            """
            if self.data_thread.is_alive():
                self.stop()

            self.stop_event.clear()
            self.data_thread = threading.Thread(target=_safe_thread_target(data_source),
                                                args=(self.data_queue, self.stop_event),
                                                kwargs=data_source_kwargs)
            self.data_thread.start()

    def __init__(self, data_source: DataSourceProtocol, *, update_interval_ms: int = int(1e3 / 60),
                 autoscale: str | None = None, autoscale_interval_ms: int | None = 1000,
                 show_fps: bool = False, useblit: bool = True, blocking_queue: bool = True,
                 event_source: TimerT | None = None, xlabel: str = '', ylabel: str = '',
                 units: Mapping[Literal['x', 'y'], str] | None = None,
                 xlim: tuple[float, float] | None = None, ylim: tuple[float, float] | None = None,
                 xscale: str | ScaleT = 'linear', yscale: str | ScaleT = 'linear',
                 style: StyleT | Sequence[StyleT] = 'fast', fig: Figure | None = None,
                 ax: Axes | None = None, fig_kw: Mapping[str, Any] | None = None,
                 log_level: int | None = None, **data_source_kwargs):

        self._LOG = logging.getLogger(f'{__name__}.{self.__class__.__qualname__}')
        self._LOG.addFilter(_ProcessNameFilter())

        if log_level is not None:
            _setup_logging(log_level)

        tic = time.perf_counter()

        self.data_queue: queue.Queue | mp.Queue = queue.Queue(maxsize=int(blocking_queue))
        self.stop_event: threading.Event | mp.Event = threading.Event()

        self.event_sources = {}
        self.update_interval_ms = int(round(update_interval_ms))
        self.show_fps = show_fps
        self.useblit = useblit
        self.autoscale = autoscale if autoscale is not None else ''
        self.autoscale_interval_ms = (int(round(autoscale_interval_ms))
                                      if autoscale_interval_ms is not None else None)
        self.fig_kw = fig_kw if fig_kw is not None else {}
        self.units = dict.fromkeys(['x', 'y'], '')
        if units is not None:
            self.units.update(units)
        self.xlim = xlim
        self.ylim = ylim
        self.style = style if isinstance(style, Sequence) else [style]

        self._event_clock = Clock(maxlen=MAX_CLOCK_LENGTH)
        self._first_frame: bool = True
        self._frame_count: int = -1
        self._running: bool | None = None
        self._text = 'Average fps: {:.1f}\nCurrent fps: {:.1f}'
        self._xlabel_text = xlabel
        self._ylabel_text = ylabel
        if self.units['x'] and xlabel != '':
            self._xlabel_text = self._xlabel_text + ' ({{{}}}{})'.format('', self.units['x'])
        if self.units['y'] and ylabel != '':
            self._ylabel_text = self._ylabel_text + ' ({{{}}}{})'.format('', self.units['y'])
        self._xscale = xscale
        self._yscale = yscale
        self._on_button_color = 'white' if 'dark_background' in self.style else 'black'
        self._off_button_color = 'tab:gray'

        with (plt.style.context(self.style, after_reset=True)):
            if fig is None:
                if ax is None:
                    fig = plt.figure(**self.fig_kw)
                else:
                    fig = ax.figure
            self.fig = fig

            assert_interactive_figure(self.fig)

            self._add_axes(ax)
            self._add_buttons()
            self._add_animated_artists()
            self._add_static_artists()

            # Attach the data source, adding interactive elements and event sources
            self.attach(data_source, event_source, **data_source_kwargs)

            self._add_events()
            self._add_event_sources(event_source)

            # We can try
            with misc.filter_warnings('ignore', UserWarning):
                try:
                    self.fig.tight_layout()
                except RuntimeError:
                    # Different layout engine
                    pass

        toc = time.perf_counter()
        self._LOG.info(f'Took {toc - tic:.2g}s to initialize.')

    @abc.abstractmethod
    def _update_labels(self, axis):
        """Reformat the axis ticks and possibly labels."""
        pass

    def _add_axes(self, ax_main: Axes | None = None):
        """Override and call to add additional axes."""
        # Another view might've already added the dedicated axes, so check first
        if ax_main is None and (ax_main := find_artist_with_label(self.fig.axes, 'main')) is None:
            ax_main = self.fig.add_subplot(label='main')
        elif ax_main.figure is not self.fig:
            raise ValueError('ax is not part of fig')
        if (ax_pause_resume := find_artist_with_label(self.fig.axes, 'pause_resume')) is None:
            ax_pause_resume = self.fig.add_axes((0.025, 0.025, 0.04, 0.04), frameon=False,
                                                label='pause_resume')
        if (ax_stop := find_artist_with_label(self.fig.axes, 'stop')) is None:
            ax_stop = self.fig.add_axes((0.075, 0.025, 0.04, 0.04), frameon=False, label='stop')
        if (ax_rescale := find_artist_with_label(self.fig.axes, 'rescale')) is None:
            ax_rescale = self.fig.add_axes((0.95, 0.025, 0.04, 0.04), frameon=False,
                                           label='rescale')

        ax_main.set_xscale(self._xscale)
        ax_main.set_yscale(self._yscale)

        if self.xlim is not None:
            ax_main.set_xlim(self.xlim)
        if self.ylim is not None:
            ax_main.set_ylim(self.ylim)

        self._xlabel_axes = ax_main
        self._ylabel_axes = ax_main

        self.axes = {'main': ax_main, 'pause_resume': ax_pause_resume, 'stop': ax_stop,
                     'rescale': ax_rescale}

    def _add_animated_artists(self):
        """Override and call to add additional animated artists."""
        self.animated_artists = {}

        if self.show_fps:
            fps = self.axes['main'].text(0.02, 0.02, '', animated=self.useblit,
                                         transform=self.axes['main'].transAxes)
            self.animated_artists['fps'] = fps

    def _add_static_artists(self):
        """Override and call to add additional static artists."""

        xlabel = self._xlabel_axes.set_xlabel(self._xlabel_text.format(''))
        ylabel = self._ylabel_axes.set_ylabel(self._ylabel_text.format(''))

        self.static_artists = {'xlabel': xlabel, 'ylabel': ylabel}

        ax = self.axes['pause_resume']
        # Another view might've already added the patches
        patches = find_artists_with_labels(ax.patches, {'resume', 'pause_1', 'pause_2'}).values()

        if all(patch is None for patch in patches):
            resume = mpatches.Polygon([[-10, -7], [1, 0], [-10, 7]],
                                      closed=True, color=self._off_button_color, label='resume')
            pause_1 = mpatches.Rectangle((3, -6), 2, 12, color=self._on_button_color,
                                         label='pause_1')
            pause_2 = mpatches.Rectangle((8, -6), 2, 12, color=self._on_button_color,
                                         label='pause_2')

            ax.add_patch(resume)
            ax.add_patch(pause_1)
            ax.add_patch(pause_2)

            ax.set_xlim(-10, 10)
            ax.set_ylim(-7, 7)
            ax.set_aspect(1)

            patches = (resume, pause_1, pause_2)

        self.static_artists['pause_resume'] = tuple(patches)

        ax = self.axes['stop']
        patches = find_artists_with_labels(ax.patches, {'stop'}).values()
        if all(patch is None for patch in patches):
            stop = mpatches.Rectangle((-7, -7), 14, 14, color=self._off_button_color,
                                      label='stop')

            ax.add_patch(stop)
            ax.set_xlim(-9, 9)
            ax.set_ylim(-9, 9)
            ax.set_aspect(1)

            patches = (stop,)

        self.static_artists['stop'] = tuple(patches)

        ax = self.axes['rescale']
        patches = find_artists_with_labels(ax.patches, {'up_arrow', 'down_arrow'}).values()

        if all(patch is None for patch in patches):
            up_arrow = mpatches.FancyArrow(0, 0, 0, 7, width=1.0, head_width=4, head_length=4,
                                           color=self._on_button_color, length_includes_head=True,
                                           label='up_arrow')
            down_arrow = mpatches.FancyArrow(0, 0, 0, -7, width=1.0, head_width=4, head_length=4,
                                             color=self._on_button_color,
                                             length_includes_head=True, label='down_arrow')
            ax.add_patch(up_arrow)
            ax.add_patch(down_arrow)

            ax.set_xlim(-10, 10)
            ax.set_ylim(-7, 7)
            ax.set_aspect(1)

            patches = (up_arrow, down_arrow)

        self.static_artists['rescale'] = tuple(patches)

    def _add_buttons(self):
        """Override and call to add additional buttons."""
        button_pause_resume = Button(self.axes['pause_resume'], '', useblit=self.useblit)
        button_stop = Button(self.axes['stop'], '', useblit=self.useblit)
        button_rescale = Button(self.axes['rescale'], '', useblit=self.useblit)
        self.buttons = {'pause_resume': button_pause_resume, 'stop': button_stop,
                        'rescale': button_rescale}

    def _add_events(self):
        """Override and call to add additional interactive events."""

        # Wrap methods that have buttons connected to fire on keys so that they
        # work with both and user code.
        def on_space(event: KeyEvent):
            if event.key == ' ':
                return self.toggle_pause_resume(event)

        def on_r(event: KeyEvent):
            if event.key.lower() == 'r':
                return self.rescale(event)

        cid_on_close = self.fig.canvas.mpl_connect('close_event', self._on_close)
        cid_on_space = self.fig.canvas.mpl_connect('key_press_event', on_space)  # type: ignore[arg-type]
        cid_on_r = self.fig.canvas.mpl_connect('key_press_event', on_r)  # type: ignore[arg-type]
        cid_button_pause_resume = self.buttons['pause_resume'].on_clicked(self.toggle_pause_resume)
        cid_button_stop = self.buttons['stop'].on_clicked(self.stop)
        cid_button_rescale = self.buttons['rescale'].on_clicked(self.rescale)
        self.cids = {'on_close': cid_on_close, 'on_space': cid_on_space, 'on_r': cid_on_r,
                     'pause_resume': cid_button_pause_resume, 'stop': cid_button_stop,
                     'rescale': cid_button_rescale}

    def _add_event_sources(self, default_event_source: TimerT | None = None):
        """Override and call to add additional event sources."""
        if default_event_source is None:
            default_event_source = self.fig.canvas.new_timer()

        default_event_source.interval = self.update_interval_ms
        default_event_source.add_callback(self._event_clock)
        default_event_source.add_callback(self._increment_frame_count)
        self.event_sources = {'default': default_event_source}

        if self.autoscale and self.autoscale_interval_ms is not None:
            # If autoscale_interval_ms is None, rescaling is done in _update()
            autoscale_event_source = self.fig.canvas.new_timer(self.autoscale_interval_ms)
            autoscale_event_source.add_callback(self.rescale, None, self.autoscale)
            self.event_sources['autoscale'] = autoscale_event_source

    def _update(self, frame) -> set[ArtistT]:
        """Called at each interval of the animation.

        Override and call to update animated artists with new data.

        There is a flag :attr:`_first_frame` which can be used to run
        code only on the very first frame. To use this in a subclass,
        be sure to call this method only afterward as the flag is reset
        here.
        """
        animated_artists = set()

        if self._first_frame:
            self._first_frame = False

            self.rescale()
            animated_artists.update(self.animated_artists.values())
        else:
            if self.show_fps:
                self.animated_artists['fps'].set_text(self._text.format(
                    1 / np.diff(self._event_clock).mean(),
                    1 / (self._event_clock[-1] - self._event_clock[-2])
                ))
                animated_artists.add(self.animated_artists['fps'])

            if self.autoscale is not None and self.autoscale_interval_ms is None:
                self.rescale(axis=self.autoscale)
                animated_artists.update(self.animated_artists.values())

        return animated_artists

    def _emit(self):
        """Queried every :attr:`update_interval_ms` ms to yield a new
        data item.

        Should not need to be overridden by subclasses.
        """
        # IMPORTANT: exception handler in the thread target sets stop event so
        # need to check the flag as otherwise would run forever.
        while not self.stop_event.is_set():
            # We check if one or more items have been put in the queue. If so,
            # yield it, but if not, spin the event loop for a short amount of
            # time to keep the figure responsive. If we didn't do this, the
            # figure would freeze while this function was waiting for a new
            # item.
            if not self.data_queue.empty():
                yield self.data_queue.get()
            else:
                self.fig.canvas.start_event_loop(EVENT_LOOP_TIMEOUT)

    def _initialize(self) -> set[ArtistT]:
        """Run once before the animation is started.

        Resize events of the figure will also trigger this method to
        run if blitting is enabled.

        Override and call to initialize elements of the plot.
        """
        self._event_clock.clear()
        self._event_clock()
        self._first_frame = True
        self._frame_count = -1

        return set(self.animated_artists.values())

    def _on_close(self, event=None):
        """Executed when the figure is closed.

        Override and call to add custom cleanup tasks.
        """
        self.stop(event)

    def _increment_frame_count(self):
        """Tracks the total number of frames."""
        self._frame_count += 1

    def _start_event_sources(self):
        for event_source in [event_source for key, event_source in self.event_sources.items()
                             if key != 'default']:
            try:
                event_source.start()
            except AttributeError:
                # Not a TimerBase event source
                pass

    def _stop_event_sources(self):
        for event_source in [source for key, source in self.event_sources.items()
                             if key != 'default']:
            try:
                event_source.stop()
            except AttributeError:
                # Not a TimerBase event source
                pass

    def _disconnect_buttons(self):
        for name, button in self.buttons.items():
            try:
                button.disconnect(self.cids[name])
            except KeyError:
                self._LOG.warning(f'The CID for button {name} was not registered in {self.cids}')
            else:
                self.cids.pop(name)

    def _disconnect_callbacks(self):
        retained = {}
        for name, cid in self.cids.items():
            if name in self.buttons or name == 'on_close':
                # handled by _disconnect_buttons / matplotlib
                retained[name] = cid
            self.fig.canvas.mpl_disconnect(cid)
        self.cids = retained

    def _set_pause_resume_state(self, state: Literal['play', 'pause']):
        play, pause_1, pause_2 = self.static_artists['pause_resume']
        if state == 'play':
            play.set_color(self._on_button_color)
            pause_1.set_color(self._off_button_color)
            pause_2.set_color(self._off_button_color)
            self._running = True
        else:
            play.set_color(self._off_button_color)
            pause_1.set_color(self._on_button_color)
            pause_2.set_color(self._on_button_color)
            self._running = False

        self.fig.canvas.draw_idle()

    def _set_stop_start_state(self, state: Literal['stop', 'start']):
        stop, = self.static_artists['stop']
        if state == 'stop':
            stop.set_color(self._on_button_color)
            self._running = None
        else:
            stop.set_color(self._off_button_color)

    @classmethod
    def in_process(cls, data_source: DataSourceProtocol, *,
                   context_method: str | None = None, manager: SyncManager | None = None,
                   backend: str | None = 'qtagg', log_level: int = logging.WARNING,
                   **kwargs) -> LiveViewProxy:
        """Run a LiveView in a separate process.

        Parameters
        ----------
        data_source :
            The data source. It is run in a separate thread of the main
            process. See the class docstring for more information.
        context_method :
            The multiprocessing context method. See
            :func:~multiprocessing.get_context`. On Linux, using 'fork'
            might make problems because it inherits the main process's
            matplotlib backend. Experimentally, using ``%matplotlib qt``
            together with 'fork' does not show the figure window.
        manager :
            A :class:`~multiprocessing.managers.SyncManager` instance to
            use for creating shared resources.
        backend :
            The matplotlib backend to use. Use None for the main process
            backend.
        log_level :
            Level for logging in the child process.
        **kwargs :
            Keyword arguments forwarded to the LiveView constructor.

        Returns
        -------
        process :
            The process running the view.

        """
        # TODO: Is serialization a performance bottleneck? If so, could share a
        #       list for array shapes / type code and use a SharedMemoryManager
        #       to share data buffers.

        def start_thread_on_condition(cv: mp.Condition, event: mp.Event, thread: threading.Thread):
            """Starts the data thread upon receiving the go-ahead from
            the plot process."""
            with cv:
                _LOG.debug(f'Waiting for {event}')
                cv.wait_for(event.is_set)

            _LOG.debug(f'Starting {thread}')
            thread.start()

        _setup_logging(log_level)

        tic = time.perf_counter()

        ctx = mp.get_context(context_method)
        if manager is None:
            # Use un-managed objects from multiprocessing the context.
            manager = ctx
        if backend is None:
            backend = matplotlib.get_backend()

        data_source_kwargs, live_view_kwargs = functools.filter_kwargs(data_source, kwargs)

        # Set up the thread that puts data in a queue in the main process
        data_queue = manager.Queue(maxsize=int(kwargs.get('blocking_queue', True)))
        stop_event = manager.Event()
        data_thread = threading.Thread(target=data_source, args=(data_queue, stop_event),
                                       kwargs=data_source_kwargs)

        # Set up a waiter thread that will start the data thread on a signal
        # from a thread started in the child process.
        condition = manager.Condition()
        start_event = manager.Event()
        waiter_thread = threading.Thread(target=start_thread_on_condition,
                                         args=(condition, start_event, data_thread))

        _LOG.debug(f'Starting {waiter_thread}')
        waiter_thread.start()

        if live_view_kwargs.pop('fig', None) is not None:
            _LOG.warning('Dropping fig kwarg for view in separate process')
        if live_view_kwargs.pop('ax', None) is not None:
            _LOG.warning('Dropping ax kwarg for view in separate process')
        live_view_kwargs.update(cv=condition, event=start_event)

        # We temporarily overwrite the qcodes config to not start all the
        # logging etc to save import time (qcodes is only used for axis
        # formatting)
        with (
                # Cannot delete because process locks files
                tempfile.TemporaryDirectory(delete=False) as tempdir,
                io.changed_directory(tempdir)
        ):
            if (qcodes := _import_qcodes()) is not None:
                rcfile = Path.cwd() / 'qcodesrc.json'
                rcfile.write_text(json.dumps(qcodes.config.defaults))
                rcfile.with_stem('qcodesrc_schema').write_text(
                    json.dumps(qcodes.config.defaults_schema)
                )

            process = ctx.Process(target=cls._run_in_process,
                                  args=(cls, data_queue, stop_event, backend, log_level),
                                  kwargs=live_view_kwargs,
                                  daemon=True)

            _LOG.debug(f'Starting {process}')
            process.start()

        toc = time.perf_counter()
        _LOG.info(f'Took {toc - tic:.2g}s to launch process {process.pid}')

        return cls.LiveViewProxy(process, data_thread, data_queue, stop_event)

    @staticmethod
    def _run_in_process(cls: type[LiveViewT], data_queue: mp.Queue, stop_event: mp.Event,
                        backend: str, log_level: int, **kwargs):
        """The plot process target.

        This starts a LiveView instance. Since the data is produced in the main
        process, we cannot pass the data production function to the constructor.
        Instead, we pass a function that emits a signal to the main process
        that the view is ready to receive data in *data_queue*.

        """

        def emit_start_signal(*_, cv: mp.Condition, event: mp.Event):
            """Emits a signal to the main process to start the data thread."""
            with cv:
                _LOG.debug('Notifying all')
                event.set()
                cv.notify_all()

            # Sleep a bit to avoid the thread being joined before the main process
            # had time to enqueue data.
            _LOG.debug(f'Sleeping 100ms in {threading.current_thread()}')
            time.sleep(100e-3)

        _setup_logging(log_level)

        tic = time.perf_counter()
        _LOG.debug('Started process')

        # Need to select an interactive backend, obviously
        _LOG.info(f'Using backend {backend}')
        matplotlib.use(backend)

        view = cls(emit_start_signal, log_level=log_level, **kwargs)
        # Overwrite view's queue and event, which are assigned to the thread
        # that signals the data thread in the main process to start, with the
        # queue and event that are shared by the two processes
        view.data_queue = data_queue
        view.stop_event = stop_event

        _LOG.debug(f'Starting {view}')
        view.start()

        toc = time.perf_counter()
        _LOG.info(f'Took {toc - tic:.2g}s to initialize and start.')

        _LOG.debug('Showing figure.')
        # Important to use block=True, otherwise the process terminates immediately
        plt.show(block=True)

        _LOG.debug('Exited.')

    def attach(self, data_source: DataSourceProtocol, event_source: TimerT | None = None,
               **data_source_kwargs):
        """Attach a data source to this view.

        The data source will be wrapped in a thread that puts data in a
        queue. Currently only a single data source can be attached at a
        time, so if a source is already attached, it will be stopped
        first.

        See :class:`LiveViewBase` for documentation of parameters.
        """
        self.stop()
        self.stop_event.clear()
        self.thread = threading.Thread(target=data_source, args=(self.data_queue, self.stop_event),
                                       kwargs=data_source_kwargs)

    def start(self):
        """Start the live view.

        Override this method to configure additional tasks to start.
        """
        if not hasattr(self, 'thread'):
            raise RuntimeError('No data source attached.')

        self._LOG.debug(f'Starting {self.thread} from {self}')
        self.thread.start()

        # Make sure the thread is running.
        time.sleep(50e-3)
        if not self.thread.is_alive():
            raise RuntimeError('thread could not be started.')

        self.animation = FuncAnimation(
            self.fig, self._update, self._emit, self._initialize,
            interval=self.update_interval_ms,
            blit=self.useblit,
            cache_frame_data=False,
            repeat=False,
            event_source=self.event_sources['default']
        )
        self._start_event_sources()
        self._set_pause_resume_state('play')
        self._set_stop_start_state('start')

    def resume(self, event=None):
        """Resume a paused live view."""
        if self._running is None:
            return self.start()
        self._set_pause_resume_state('play')
        self._start_event_sources()
        self.animation.resume()

    def pause(self, event=None):
        """Pause the live view."""
        if self._running is None:
            return
        self.animation.pause()
        self._stop_event_sources()
        self._set_pause_resume_state('pause')

    def stop(self, event=None):
        """Terminate the live view and interactive elements."""
        self._LOG.debug(f'Stopping view on event {event}.')

        if self._running is None:
            return

        # Set UI state
        self.pause(event)
        self._set_stop_start_state('stop')
        if event is None:
            # handled by Animation's figure close hook else.
            self.animation.stop()

        # Stop data sources
        self.stop_event.set()
        # Need to fetch one last time in case the thread that puts data is
        # waiting for the queue to be free, preventing it from joining once
        # it encounters the stop flag
        while not self.data_queue.empty():
            self.data_queue.get(block=False)
        try:
            self.thread.join(timeout=1)
        except RuntimeError:
            # Thread not started
            pass

    def toggle_pause_resume(self, event=None):
        """Pause/Resume the live view."""
        if self._running is None:
            self.start()
        elif self._running:
            # Pauses 'default' event source
            self.pause()
        else:
            self.resume()

    @_queue_draw
    def rescale(self, event=None, axis: Literal['', 'x', 'y', 'xy'] = 'xy') -> bool:
        """Rescale the main axis.

        This is only actually performed if :attr:`xlim` or :attr:`ylim`
        are not overridden to return not None.

        If subclass implementations do not choose to call this method
        they should be decorated with :func:`_queue_draw` to ensure the
        figure updates.

        Parameters
        ----------
        event :
            Only required for matplotlib internals.
        axis :
            The axis to rescale. Can be any combination of 'x' and 'y'.

        Returns
        -------
        draw_required :
            Boolean indicating that a complete figure redraw is
            required. Handled by the :func:`_queue_draw` decorator.

        """
        # Only rescale where no static limits are defined by overriding the properties
        if 'x' in axis and self.xlim is not None:
            axis = axis.replace('x', '')
        if 'y' in axis and self.ylim is not None:
            axis = axis.replace('y', '')
        if axis == '':
            return False

        # TODO: necessary?
        self.axes['main'].relim()
        self.axes['main'].autoscale_view(scalex='x' in axis, scaley='y' in axis)

        self._update_labels(axis)

        return True


class LiveView1DBase(LiveViewBase, abc.ABC):
    r"""Abstract base class for 1d live views.

    Subclasses implement :meth:`_update_ydata` to define how the 1d
    y-data is updated in each frame. y-data can be returned by the
    `data_source` in the following ways:

        - a 1d array of shape (x.size,), in which case a single line
          will be plotted.
        - a 2d array of shape (x.size, n_lines), in which case
          `n_lines` lines will be plotted.
        - a dictionary ``{label: 1d array}`` in which case a line will
          be drawn for each entry in the dictionary (as long as
          `n_lines` matches).

    Parameters
    ----------
    n_lines :
        The number of lines to plot. Should correspond to y-data's
        second dimension if an array or the size of the dictionary
        else.
    plot_legend :
        Plot a legend. If a string, passed through to
        :meth:`Axes.legend`\ s `loc` parameter. To have meaningful
        line labels, `data_source` should return a dictionary as its
        second return value whose keys will be taken as labels.

    For further parameters, see the base class.

    """
    _ydata: npt.NDArray[np.float64]
    """The data buffer populated at each frame by :meth:`_update_ydata."""

    def __init__(self, data_source: DataSourceProtocol, *,
                 n_lines: int = 1, plot_legend: bool | str = False,
                 update_interval_ms: int = int(1e3 / 60), autoscale: str | None = None,
                 autoscale_interval_ms: int | None = 1000, show_fps: bool = False,
                 useblit: bool = True, blocking_queue: bool = True,
                 event_source: TimerT | None = None, xlabel: str = '', ylabel: str = '',
                 units: Mapping[Literal['x', 'y'], str] | None = None,
                 xlim: tuple[float, float] | None = None, ylim: tuple[float, float] | None = None,
                 xscale: str | ScaleT = 'linear', yscale: str | ScaleT = 'linear',
                 style: StyleT | Sequence[StyleT] = 'fast', fig: Figure | None = None,
                 ax: Axes | None = None, fig_kw: Mapping[str, Any] | None = None,
                 log_level: int | None = None, **data_source_kwargs):
        self.n_lines = n_lines
        self.plot_legend = plot_legend
        super().__init__(data_source, update_interval_ms=update_interval_ms, autoscale=autoscale,
                         autoscale_interval_ms=autoscale_interval_ms, show_fps=show_fps,
                         useblit=useblit, blocking_queue=blocking_queue, event_source=event_source,
                         xlabel=xlabel, ylabel=ylabel, units=units, xlim=xlim, ylim=ylim,
                         xscale=xscale, yscale=yscale, style=style, fig=fig, ax=ax, fig_kw=fig_kw,
                         log_level=log_level, **data_source_kwargs)

    @abc.abstractmethod
    def _update_ydata(self, frame):
        """Run before updating the line's ydata with :attr:`_ydata`.

        Subclasses should update that array or replace it with data
        from `frame` as needed.
        """
        pass

    @property
    def _line_handles(self) -> list[ArtistT]:
        return [artist for key, artist in self.animated_artists.items() if key.startswith('line')]

    def _add_animated_artists(self):
        super()._add_animated_artists()
        for i in range(self.n_lines):
            line, = self.axes['main'].plot([], [], animated=self.useblit)
            self.animated_artists[f'line_{i}'] = line

    def _add_static_artists(self):
        super()._add_static_artists()
        if self.plot_legend:
            loc = self.plot_legend if isinstance(self.plot_legend, str) else 'best'
            self.static_artists['legend'] = self.axes['main'].legend(
                handles=self._line_handles,
                labels=[''] * len(self._line_handles),
                loc=loc,
                framealpha=1
            )

    def _update(self, frame: tuple) -> set[ArtistT]:
        x, y = frame

        if self._first_frame:
            self._ydata = np.full((len(x), self.n_lines), np.nan)
            if self.plot_legend and isinstance(y, Mapping):
                for txt, key in zip(self.static_artists['legend'].texts, y):
                    txt.set_text(key)

        # Update data before call to super's method so that artist extents are known for
        # rescaling e.g.
        self._update_ydata(frame)
        # Usually, x will not change, so we could use set_ydata(). But it doesn't
        # hurt to allow the possibility of x being dynamic.
        for i in range(self.n_lines):
            self.animated_artists[f'line_{i}'].set_data(x, self._ydata[:, i])

        animated_artists = super()._update(frame)
        animated_artists.update(self._line_handles)

        return animated_artists

    def _update_labels(self, axis):
        if _import_qcodes() is None:
            # reformat_axis() uses qcodes
            return

        if 'x' in axis and self._xlabel_text != '':
            prefix = reformat_axis(self.axes['main'],
                                   [line.get_xdata() for line in self._line_handles],
                                   self.units['x'], which='x')
            self.static_artists['xlabel'].set_text(self._xlabel_text.format(prefix))
        if 'y' in axis and self._ylabel_text != '':
            prefix = reformat_axis(self.axes['main'],
                                   [line.get_ydata() for line in self._line_handles],
                                   self.units['y'], which='y')
            self.static_artists['ylabel'].set_text(self._ylabel_text.format(prefix))


class BatchedLiveView1D(LiveView1DBase):
    """A live view for batched 1d data.

    `data_source` should produce a two-tuple of (x, y) data for this
    view, where x is a 1d array and y is

     - a 1d array of shape ``(x.size,)``
     - a 2d array of shape ``(x.size, n_lines)``
     - a dictionary ``{label: 1d array}``

    For parameters, see :class:`LiveView1DBase`.

    Examples
    --------
    Define a function that generates random data. The function will be
    pushed to a background thread so that the interpreter is unblocked.

    >>> def produce_data(interval=1e-1):
    ...     import time  # imports for tests in child process
    ...     import numpy as np
    ...     rng = np.random.default_rng()
    ...     t0 = t1 = time.perf_counter()
    ...     while True:
    ...         dt = time.perf_counter() - t1
    ...         if interval > dt:
    ...             time.sleep(interval - dt)
    ...         t1 = time.perf_counter()
    ...         x = 1 + (t1 - t0) * 100
    ...         yield np.arange(1024), x * rng.standard_normal(1024)

    Wrap produce_data() so that it puts its generated values into a
    queue until an event is set:

    >>> put_data = iterable_data_source_factory(produce_data)

    Instantiate the live view object:

    >>> view = BatchedLiveView1D(put_data, autoscale='y', interval=1e-3,
    ...                          show_fps=True)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Pause the update (and the data production). This can also be done
    by clicking the play/pause button in the lower left corner:

    >>> view.toggle_pause_resume()

    ... and resume again:

    >>> view.toggle_pause_resume()

    Rescaling of the data limits can also be triggered manually
    (independent of the `autoscale` setting) by clicking on the up-down
    arrow symbol in the lower left corner or programmatically:

    >>> view.rescale(axis='xy')

    Terminate the data production thread and disable interactive
    elements of the figure:

    >>> view.stop()

    Once a view is stopped, new data sources can be attached to it and
    the live view started again:

    >>> view.attach(put_data, interval=1e-3)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    >>> view.stop()

    If the data update rate is slow, blitting together with automatic
    rescaling might lead to a white canvas being visible between a
    rescale being triggered and new data being drawn. In this case you
    can disable blitting, but note that this will be less performant:

    >>> view = BatchedLiveView1D(put_data, autoscale='y', useblit=False,
    ...                          interval=0.1, show_fps=True)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Terminate the data production thread and disable interactive
    elements of the figure:

    >>> view.stop()

    The view can also be run in a separate process. This classmethod
    returns a :class:`LiveViewProxy` object which contains the
    :class:`~multiprocessing.Process` object as well as the
    :class:~multiprocessing.Queue` and :class:`~multiprocessing.Event`
    used for communication (no doctest because it hangs)::

        proxy = BatchedLiveView1D.in_process(put_data)

    The data streaming can be stopped without killing the live view::

        proxy.stop()

    Afterwards, a new data source can be attached::

        proxy.attach(put_data)

    Stop the data production thread and terminate the process::

        proxy.stop()
        proxy.process.terminate()

    Multiple lines are also supported. Define a data source that
    returns a dictionary for y-values:

    >>> def produce_data(interval=1e-1):
    ...     import time  # imports for tests in child process
    ...     import numpy as np
    ...     rng = np.random.default_rng()
    ...     t0 = t1 = time.perf_counter()
    ...     while True:
    ...         dt = time.perf_counter() - t1
    ...         if interval > dt:
    ...             time.sleep(interval - dt)
    ...         t1 = time.perf_counter()
    ...         x = 1 + (t1 - t0) * 100
    ...         y = {'foo': +x * rng.standard_normal(1024),
    ...              'bar': -x * rng.standard_normal(1024)}
    ...         yield np.arange(1024), y
    >>> put_data = iterable_data_source_factory(produce_data)
    >>> view = BatchedLiveView1D(put_data, n_lines=2, plot_legend=True,
    ...                          autoscale='y', interval=1e-3)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    >>> view.stop()

    """

    def _update_ydata(self, frame):
        x, y = frame
        # At first, we assume that the provided data buffer should be used. If
        # that fails because y has the wrong size, we replace the data buffer
        # with it.
        if isinstance(y, Mapping):
            try:
                for i, data in enumerate(y.values()):
                    self._ydata[:, i] = data
            except ValueError:
                self._ydata = np.array([val for val in y.values()]).T
        else:
            y = np.array(y)[:, None] if np.ndim(y) == 1 else y
            try:
                self._ydata[:] = y
            except ValueError:
                self._ydata = y


class IncrementalLiveView1D(LiveView1DBase):
    """A live view for 1d data arriving in single samples.

    `data_source` should produce a two-tuple of (x, y) data for this
    view, where x is a 1d array and y is

     - a number
     - a sequence of length `n_lines`
     - a dictionary ``{label: number}``.

    For parameters, see :class:`LiveView1DBase`.

    Examples
    --------
    >>> rng = np.random.default_rng()
    >>> def produce_data(interval=1e-3):
    ...     t0 = t1 = time.perf_counter()
    ...     while True:
    ...         dt = time.perf_counter() - t1
    ...         if interval > dt:
    ...             time.sleep(interval - dt)
    ...         t1 = time.perf_counter()
    ...         x = 1 + (t1 - t0) * 100
    ...         yield np.arange(128), x * rng.standard_normal()

    >>> put_data = iterable_data_source_factory(produce_data)
    >>> view = IncrementalLiveView1D(put_data, autoscale='xy', xlim=(0, 127),
    ...                              autoscale_interval_ms=None)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Terminate the data production thread and disable interactive
    elements of the figure:

    >>> view.stop()

    Multiple lines are also supported. Define a data source that
    returns a sequence of y-values:

    >>> def produce_data(interval=1e-3):
    ...     t0 = t1 = time.perf_counter()
    ...     while True:
    ...         dt = time.perf_counter() - t1
    ...         if interval > dt:
    ...             time.sleep(interval - dt)
    ...         t1 = time.perf_counter()
    ...         x = 1 + (t1 - t0) * 100
    ...         yield np.arange(1024), x * rng.standard_normal(2)
    >>> put_data = iterable_data_source_factory(produce_data)
    >>> view = IncrementalLiveView1D(put_data, n_lines=2, autoscale='xy',
    ...                              autoscale_interval_ms=None)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    >>> view.stop()
    """

    def _update_ydata(self, frame):
        x, y = frame

        if not (ix := self._frame_count % len(x)):
            # Start of new 1d trace, reset buffer to nans
            self._ydata[1:] = np.nan

        if isinstance(y, Mapping):
            for i, data in enumerate(y.values()):
                self._ydata[ix, i] = data
        else:
            self._ydata[ix] = y


class LiveView2DBase(LiveViewBase, abc.ABC):
    """Abstract base class for 2d live views.

    Subclasses implement :meth:`_update_cdata` to define how the 2d
    data is updated in each frame.

    Parameters
    ----------
    autoscale :
        Automatically rescale axes.

         - 'c': rescale the extents and colorbar of the colorplot
         - 'y': rescale the y-axis of the line plot (if any)

    plot_line :
        Show a line plot with the average along the y-axis.
    clabel :
        The label of the colorbar.
    img_kw :
        Keyword-arguments passed on to
        :meth:`matplotlib:matplotlib.axes.Axes.pcolorfast`. Use this to
        specify a norm, for instance.

    For additional parameters, see :class:`LiveViewBase`.

    """
    _cdata: npt.NDArray[np.float64]
    """The data buffer populated at each frame by :meth:`_update_cdata."""
    cbar: colorbar.Colorbar
    units: dict[Literal['x', 'y', 'c'], str]

    def __init__(self, data_source: DataSourceProtocol, update_interval_ms: int = int(1e3 / 60),
                 autoscale: Literal['', 'y', 'c', 'yc'] | None = None,
                 autoscale_interval_ms: int | None = 1000, plot_line: bool = False,
                 show_fps: bool = False, useblit: bool = True, blocking_queue: bool = True,
                 event_source: TimerT | None = None, xlabel: str = '', ylabel: str = '',
                 clabel: str = '', units: dict[Literal['x', 'y', 'c'], str] | None = None,
                 xlim: tuple[float, float] | None = None, ylim: tuple[float, float] | None = None,
                 xscale: str | ScaleT = 'linear', yscale: str | ScaleT = 'linear',
                 fig: Figure | None = None, ax: Axes | None = None,
                 img_kw: Mapping[str, Any] | None = None, fig_kw: Mapping[str, Any] | None = None,
                 log_level: int | None = None, **data_source_kwargs):

        if not (xscale == 'linear' or isinstance(xscale, scale.LinearScale)):
            self._LOG.warning('non-linear scale not yet supported by pcolorfast')
            xscale = 'linear'
        if not (yscale == 'linear' or isinstance(yscale, scale.LinearScale)):
            self._LOG.warning('non-linear scale not yet supported by pcolorfast')
            yscale = 'linear'

        # need to define these before calling super()'s init because that calls
        # methods overridden by this class which might need these attributes.
        self.plot_line = plot_line
        self.img_kw = img_kw if img_kw is not None else {}
        if 'animated' in self.img_kw:
            self._LOG.warning('img_kw `animated` ignored in favor of useblit.')
            self.img_kw.pop('animated')
        self._clabel_text = clabel
        self._linelabel_text = self.line_operation.label
        units = dict.fromkeys(['x', 'y', 'c'], '') if units is None else units
        if units.setdefault('c', ''):
            self._clabel_text = self._clabel_text + ' ({{{}}}{})'.format('', units['c'])
            self._linelabel_text = self._linelabel_text + ' ({{{}}}{})'.format('', units['c'])

        super().__init__(data_source, update_interval_ms=update_interval_ms, autoscale=autoscale,
                         autoscale_interval_ms=autoscale_interval_ms, show_fps=show_fps,
                         useblit=useblit, blocking_queue=blocking_queue, event_source=event_source,
                         xlabel=xlabel, ylabel=ylabel, units=units, xlim=xlim, ylim=ylim,
                         xscale=xscale, yscale=yscale, fig=fig, ax=ax, fig_kw=fig_kw,
                         log_level=log_level, **data_source_kwargs)

    @abc.abstractmethod
    def _update_cdata(self, frame):
        """Run before updating the image with :attr:`_cdata`.

        Subclasses should update that array with data from `frame` or
        overwrite it as needed.
        """
        pass

    @property
    @abc.abstractmethod
    def line_operation(self) -> LineOperation:
        """The operation applied to the 2d data to obtain 1d data."""
        pass

    def _add_axes(self, ax=None):
        super()._add_axes(ax)

        # Make room for colorbar
        if self.plot_line:
            gs = _make_subgridspec(self.axes['main'], height_ratios=[3, 1])

            self.axes['cbar'] = self.fig.add_subplot(gs[0, 1], label='<colorbar>')
            self.axes['cbar'].set_anchor((0.0, 0.5))
            self.axes['cbar'].set_box_aspect(20)
            self.axes['cbar'].set_aspect('auto')

            self.axes['line'] = self.fig.add_subplot(gs[1, 0], sharex=self.axes['main'],
                                                     label='line')
            self.axes['line'].yaxis.set_tick_params(which='both', left=True, right=True,
                                                    labelright=True, labelleft=False)
            self.axes['line'].yaxis.set_label_position('right')
            self.axes['line'].set_yscale(norm_to_scale(self.img_kw.get('norm')))
            self.axes['main'].label_outer()
            # Indicate that the line plot is the lowest axes to be used for labeling
            self._xlabel_axes = self.axes['line']
        else:
            self.axes['cbar'], _ = colorbar.make_axes_gridspec(self.axes['main'])

    def _add_animated_artists(self):
        super()._add_animated_artists()

        img = self.axes['main'].pcolorfast(np.full((1, 1), np.nan), animated=self.useblit,
                                           **self.img_kw)
        self.animated_artists['img'] = img

        if self.plot_line:
            line, = self.axes['line'].plot([], [], animated=self.useblit)
            self.animated_artists['line'] = line

    def _add_static_artists(self):
        super()._add_static_artists()

        self.cbar = self.fig.colorbar(self.animated_artists['img'], cax=self.axes['cbar'])
        self.cbar.set_label(self._clabel_text.format(''))

        # The correct Text object is a bit hidden
        self.static_artists['clabel'] = self.cbar.ax.yaxis.label

        if self.plot_line:
            linelabel = self.axes['line'].set_ylabel(self._linelabel_text.format(''))
            self.static_artists['linelabel'] = linelabel

    def _update(self, frame) -> set[ArtistT]:
        x, y, c = frame

        # Do this before call to super()'s because that resets the _first_frame flag
        if self._first_frame:
            self._cdata = np.full((len(y), len(x)), np.nan)

        xext, yext = itertools.split_into(self.animated_artists['img'].get_extent(), (2, 2))
        xlim, ylim = [x[0], x[-1]], [y[0], y[-1]]
        if self.autoscale and (xext != xlim or yext != ylim):
            # set_extent() deviates in behavior from set_data() in that it automatically
            # relims. Hence, we need to deviate from the design and enqueue the draw here
            # instead of letting rescale() handle it.
            self.animated_artists['img'].set_extent(xlim + ylim)
            self._update_labels('x')
            self.animation.draw_enqueued.set()

        # Update data before call to super's method so that artist extents are known for
        # rescaling e.g.
        self._update_cdata(frame)
        self.animated_artists['img'].set_data(self._cdata)
        if self.plot_line:
            self.animated_artists['line'].set_data(x, self.line_operation(self._cdata))

        animated_artists = super()._update(frame)
        animated_artists.add(self.animated_artists['img'])
        if self.plot_line:
            animated_artists.add(self.animated_artists['line'])

        return animated_artists

    def _update_labels(self, axis):
        if _import_qcodes() is None:
            # reformat_axis() uses qcodes
            return

        if 'x' in axis:
            # TODO: this is awkward. 'x' means x and y labels of the main plot
            if self._xlabel_text != '':
                prefix = reformat_axis(self.axes['main'],
                                       self.animated_artists['img'].get_extent()[:2],
                                       self.units['x'], which='x')
                self.static_artists['xlabel'].set_text(self._xlabel_text.format(prefix))
            if self._ylabel_text != '':
                prefix = reformat_axis(self.axes['main'],
                                       self.animated_artists['img'].get_extent()[2:],
                                       self.units['y'], which='y')
                self.static_artists['ylabel'].set_text(self._ylabel_text.format(prefix))
        if 'c' in axis and self._clabel_text != '':
            # colorbar
            prefix = reformat_axis(self.cbar,
                                   self.animated_artists['img'].get_array(),
                                   self.units['c'], which='c')
            self.static_artists['clabel'].set_text(self._clabel_text.format(prefix))
        if 'y' in axis and self.plot_line and self._linelabel_text != '':
            # 'y' means the line plot
            prefix = reformat_axis(self.axes['line'], self.animated_artists['line'].get_ydata(),
                                   self.units['c'], which='y')
            self.static_artists['linelabel'].set_text(self._linelabel_text.format(prefix))

    @_queue_draw
    def rescale(self, event=None, axis: Literal['', 'y', 'c', 'yc'] = 'yc') -> bool:
        draw_required = False

        vmin = self.img_kw.get('vmin', False)
        vmax = self.img_kw.get('vmax', False)
        if 'c' in axis and (vmin is False or vmax is False):
            # At least one limit dynamic, update
            with self.animated_artists['img'].norm.callbacks.blocked():
                # Do not use img.autoscale because vmin/max might be given which it ignores
                if vmin is False:
                    self.animated_artists['img'].norm.vmin = None
                if vmax is False:
                    self.animated_artists['img'].norm.vmax = None

            self.animated_artists['img'].autoscale_None()
            draw_required = True

        if 'y' in axis and self.plot_line:
            # 'x' case is controlled by _update()
            # TODO: necessary?
            self.axes['line'].relim()
            self.axes['line'].autoscale_view(scalex=False, scaley='y' in axis)
            draw_required = True

        self._update_labels(axis)

        return draw_required


class BatchedLiveView2D(LiveView2DBase):
    """A live view for 2d batched data.

    `data_source` should produce a three-tuple of (x, y, c) data for
    this view, where x and y are 1 or 2d arrays and c is a 2d array
    of shape (y_pts, x_pts).

    This class does *not* update x and y extents of the image.

    For parameters, see :class:`LiveView2DBase`.

    Examples
    --------
    >>> rng = np.random.default_rng()
    >>> def produce_data(interval=1e-1):
    ...     t0 = t1 = time.perf_counter()
    ...     x, y = np.linspace(3.2, 5.7, 1920), np.linspace(0, 1, 1080)
    ...     while True:
    ...         dt = time.perf_counter() - t1
    ...         if interval > dt:
    ...             time.sleep(interval - dt)
    ...         t1 = time.perf_counter()
    ...         c = 1 + (t1 - t0) * 100
    ...         yield x, y, c * rng.standard_normal((1080, 1920))

    >>> put_data = iterable_data_source_factory(produce_data)
    >>> view = BatchedLiveView2D(put_data, autoscale='c', interval=1e-3)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Stop the event loop:

    >>> view.stop()

    Plot the mean of 1d traces along the y axis as a line plot:

    >>> view = BatchedLiveView2D(put_data, autoscale='yc', interval=1e-3,
    ...                          plot_line=True)
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Stop the event loop:

    >>> view.stop()

    """

    @functools.cached_property
    def line_operation(self) -> LineOperation:
        def op(data: npt.ArrayLike) -> npt.NDArray:
            return np.nanmean(data, axis=0)

        op.label = 'Mean'
        return op

    def _update_cdata(self, frame):
        x, y, c = frame
        # At first, we assume that the provided data buffer should be used. If
        # that fails because c has the wrong size, we replace the data buffer
        # with it.
        try:
            self._cdata[:] = c
        except ValueError:
            self._cdata = c


class IncrementalLiveView2D(LiveView2DBase):
    """A live view for 2d data arriving in 1d batches.

    The x-axis is assumed to be the fast axis, meaning the 2d image is
    filled in row-by-row using incoming 1d data batches.

    `data_source` should produce a three-tuple of (x, y, c) data for
    this view, where `c` is 1d with ``len(c) == len(x)``.

    For parameters, see :class:`LiveView2DBase`.

    Examples
    --------
    >>> rng = np.random.default_rng()
    >>> def produce_data(interval=1e-3):
    ...     t0 = t1 = time.perf_counter()
    ...     x, y = np.linspace(3.2, 5.7, 200), np.linspace(0, 1, 20)
    ...     while True:
    ...         dt = time.perf_counter() - t1
    ...         if interval > dt:
    ...             time.sleep(interval - dt)
    ...         t1 = time.perf_counter()
    ...         c = 1 + (t1 - t0) * 100
    ...         yield x, y, c * rng.standard_normal(200)

    >>> put_data = iterable_data_source_factory(produce_data)
    >>> view = IncrementalLiveView2D(put_data, autoscale='c')
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Stop the event loop:

    >>> view.stop()

    Plot the last acquired 1d trace as well:

    >>> view = IncrementalLiveView2D(put_data, autoscale='yc', plot_line=True,
    ...                              clabel='my label')
    >>> view.start()
    >>> plt.pause(0.1)  # Only here for doctest

    Stop the event loop:

    >>> view.stop()

    """

    @functools.cached_property
    def line_operation(self) -> LineOperation:
        def op(data: npt.ArrayLike) -> npt.NDArray:
            return data[self._frame_count % self._cdata.shape[0]]

        op.label = self._clabel_text
        return op

    def _update_cdata(self, frame):
        x, y, c = frame

        if not (ix := self._frame_count % len(y)):
            # Start of new 2d image, reset buffer to nans
            self._cdata[1:] = np.nan

        self._cdata[ix] = c


def _make_subgridspec(ax: Axes, height_ratios: Sequence[float], pad: float = 0.05,
                      fraction: float = 0.15, hspace: float = 0.1) -> GridSpecFromSubplotSpec:
    # Copied from matplotlib.colorbar.make_axes_gridspec()
    wh_space = 2 * pad / (1 - pad)

    gs = ax.get_subplotspec().subgridspec(
        nrows=len(height_ratios), ncols=2, wspace=wh_space, hspace=hspace,
        height_ratios=height_ratios, width_ratios=[1 - fraction - pad, fraction]
    )
    ax.set_subplotspec(gs[0, 0])
    return gs


def with_interval_once(event_source: TimerT, temporary_interval_ms: int = 0) -> TimerT:
    """Set `event_source`\'s interval to `temporary_interval_ms` once
    before resetting it.

    Parameters
    ----------
    event_source :
        The even source whose interval to adjust.
    temporary_interval_ms : int, optional
        The interval to be used just once. The default is 0.

    Returns
    -------
    TimerT
        The event source itself. Same object as the input argument.

    """

    def reset(interval_ms: int):
        event_source.remove_callback(reset)
        event_source.interval = interval_ms

    event_source.add_callback(reset, event_source.interval)
    event_source.interval = temporary_interval_ms
    return event_source


def iterable_data_source_factory(
        iterable_function: Callable[P.kwargs, Iterable[T]]
) -> DataSourceProtocol[P, T]:
    """Return a function that puts data from the iterable returned by
    `iterable_function` into a queue until a stop flag is set."""

    def data_source(data_queue: queue.Queue[T], stop_event: threading.Event, **kwargs: P.kwargs):
        it = iter(iterable_function(**kwargs))
        while not stop_event.is_set():
            data_queue.put(next(it))

    # Update the signature so that it can be inspected.
    iterable_signature = inspect.signature(iterable_function)
    original_signature = inspect.signature(data_source)
    data_source = functools.wraps(iterable_function)(data_source)

    data_source.__signature__ = inspect.signature(data_source).replace(
        parameters=[original_signature.parameters['data_queue'],
                    original_signature.parameters['stop_event'],
                    *iterable_signature.parameters.values()]
    )

    return data_source


def find_artist_with_label(artists: Iterable[ArtistT], label: str) -> Artist | None:
    """Find the artist, if any, whose label is `label`.

    Returns None otherwise.
    """
    return itertools.first_true(artists, pred=lambda artist: artist.get_label() == label)


def find_artists_with_labels(artists: Iterable[ArtistT],
                             labels: Iterable[str]) -> dict[str, ArtistT | None]:
    """Find the artists, if any, whose labels are in `labels`.

    Returns a dictionary with keys `labels` and the artist if found and
    None otherwise as values.
    """
    matches = dict.fromkeys(labels)
    for artist in artists:
        if (label := artist.get_label()) in labels:
            matches[label] = artist

    return matches


def _safe_thread_target(func):
    @functools.wraps(func)
    def wrapped(data_queue: queue.Queue[T], stop_event: threading.Event, **kwargs):
        try:
            func(data_queue, stop_event, **kwargs)
        except Exception as exc:
            _LOG.exception('Caught an exception in data source. Quitting gracefully.',
                           exc_info=exc)
            stop_event.set()
            return

    return wrapped


def _import_qcodes() -> ModuleType | None:
    if (module := sys.modules.get('qcodes')) is not None:
        return module
    try:
        return importlib.import_module('qcodes')
    except ImportError:
        return None


class _ProcessNameFilter(logging.Filter):
    def filter(self, record):
        record.processname = 'process '
        if (pid := os.getpid()) == 1:
            record.processname += '(initial)'
        else:
            record.processname += str(pid)
        if mp.parent_process() is None:
            record.processname += ' (main)'

        return True


def _setup_logging(level: int):
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(processname)s - %(message)s'
    )

    root_logger = logging.getLogger('qutil')
    root_logger.setLevel(logging.WARNING)
    root_logger.propagate = False

    logger = logging.getLogger(__name__)
    logger.setLevel(level)
    logger.addFilter(_ProcessNameFilter())

    for handler in logger.handlers:
        handler.close()
        logger.removeHandler(handler)

    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)

    logger.addHandler(handler)


LiveViewT = TypeVar('LiveViewT', bound=LiveViewBase)



from __future__ import annotations

import inspect
import logging
import sys
import warnings
from collections.abc import Callable, MutableMapping
from contextlib import contextmanager
from importlib import import_module
from types import ModuleType
from typing import Dict, Optional, Union, Any, Hashable, TypeVar
from unittest.mock import MagicMock


@contextmanager
def all_logging_disabled(highest_level=logging.CRITICAL):
    """
    A context manager that will prevent any logging messages
    triggered during the body from being processed.
    :param highest_level: the maximum logging level in use.
    This would only need to be changed if a custom level greater than CRITICAL
    is defined.

    https://gist.github.com/simon-weber/7853144
    """
    # two kind-of hacks here:
    #    * can't get the highest logging level in effect => delegate to the user
    #    * can't get the current module-level override => use an undocumented
    #       (but non-private!) interface

    previous_level = logging.root.manager.disable

    logging.disable(highest_level)

    try:
        yield
    finally:
        logging.disable(previous_level)


@contextmanager
def filter_warnings(action, category=Warning, lineno=0, append=False, *,
                    record=False, module=None):
    """A context manager that combines catching and filtering warnings."""
    with warnings.catch_warnings(record=record, module=module) as manager:
        warnings.simplefilter(action, category, lineno, append)
        try:
            yield manager
        finally:
            pass


@contextmanager
def key_set_to(dct: MutableMapping, key: Hashable, val: Any):
    """Temporarily set `key` in `dct` to `val`.

    Examples
    --------
    >>> my_dict = {'a': 2, 3: 'b'}
    >>> my_dict['a']
    2
    >>> with key_set_to(my_dict, 'a', 3):
    ...     print(my_dict['a'])
    3
    >>> my_dict['a']
    2

    Also works with previously nonexisting keys:

    >>> with key_set_to(my_dict, 1, 2):
    ...     print(my_dict[1])
    2
    >>> 1 in my_dict
    False

    """
    missing_sentinel = object()
    previous = dct.get(key, missing_sentinel)

    try:
        dct[key] = val
        yield dct
    finally:
        try:
            if previous is missing_sentinel:
                del dct[key]
            else:
                dct[key] = previous
        except:
            pass


@contextmanager
def attr_set_to(obj: Any, attr: str, val: Any, allow_missing: bool = False):
    """Temporarily set `attr` in `obj` to `val`.

    If `allow_missing` is `True`, `attr` will also be set if it did not
    exist before.

    Examples
    --------
    >>> class Foo:
    ...     a = 3
    >>> foo = Foo()
    >>> foo.a
    3
    >>> with attr_set_to(foo, 'a', 4):
    ...     print(foo.a)
    4
    >>> foo.a
    3
    >>> with attr_set_to(foo, 'b', 1, allow_missing=True):
    ...     print(foo.b)
    1
    >>> hasattr(foo, 'b')
    False

    """
    missing_sentinel = object()
    try:
        previous = getattr(obj, attr)
    except AttributeError:
        if allow_missing:
            previous = missing_sentinel
        else:
            raise

    try:
        setattr(obj, attr, val)
        yield obj
    except AttributeError:
        raise
    finally:
        try:
            if previous is missing_sentinel:
                delattr(obj, attr)
            else:
                setattr(obj, attr, previous)
        except:
            pass


def import_or_mock(
        name: str, package: Optional[str] = None, local_name: Optional[str] = None
) -> Dict[str, Union[ModuleType, MagicMock]]:
    """Imports a module or, if it cannot be imported, mocks it.

    If it is importable, equivalent to::
        from name import package as local_name

    Parameters
    ----------
    name : str
        See :func:`importlib.import_module`.
    package : str | None
        See :func:`importlib.import_module`.
    local_name : str
        Either the name assigned to the module or the object to be
        imported from the module.

    Returns
    -------
    dict[str, ModuleType | MagicMock]
        A dictionary with the single entry {local_name: module}.

    Examples
    --------
    >>> locals().update(import_or_mock('numpy', None, 'pi'))
    >>> pi
    3.141592653589793
    >>> locals().update(import_or_mock('owiejlkjlqz'))
    >>> owiejlkjlqz
    <MagicMock name='mock.owiejlkjlqz' id='...'>

    """
    local_name = local_name or name
    try:
        module = import_module(name, package)
    except ImportError:
        module = MagicMock(__name__=name)
    return {local_name: getattr(module, local_name, module)}



from . import live_view as live_view
from .core import (
    CoordClicker as CoordClicker,
    LineClicker as LineClicker,
    LaTeXPlotHelper as LaTeXPlotHelper,
    BlitManager as BlitManager,
    plot_blitted as plot_blitted,
    pcolorfast_blitted as pcolorfast_blitted,
    changed_plotting_backend as changed_plotting_backend,
    plot_2d_dataframe as plot_2d_dataframe,
    cycle_plots as cycle_plots,
    update_plot as update_plot,
    list_styles as list_styles,
    reformat_axis as reformat_axis,
    norm_to_scale as norm_to_scale,
    assert_interactive_figure as assert_interactive_figure,
    get_rwth_color_cycle as get_rwth_color_cycle,
    rwth_color_cycle as rwth_color_cycle,
    rwth_color_cycle_25 as rwth_color_cycle_25,
    rwth_color_cycle_50 as rwth_color_cycle_50,
    rwth_color_cycle_75 as rwth_color_cycle_75,
    rwth_color_cycle_100 as rwth_color_cycle_100,
    RWTH_COLORS as RWTH_COLORS
)
from .live_view import (
    BatchedLiveView1D as BatchedLiveView1D,
    BatchedLiveView2D as BatchedLiveView2D,
    IncrementalLiveView1D as IncrementalLiveView1D,
    IncrementalLiveView2D as IncrementalLiveView2D
)



from typing import List, Hashable, Union, Iterator, Tuple, Any
import pandas as pd
import numpy as np


def consecutive_groupby(df: pd.DataFrame,
                        columns: Union[Hashable, List[Hashable]]) -> Iterator[Tuple[Any, pd.DataFrame]]:
    """Iterate over the given data frame in groups where the specified columns have the same value in consecutive rows.

    Parameters
    ----------
    df
    columns

    Examples
    --------
    >>> df = pd.DataFrame([['a', 0., 42], ['a', 1., 42], ['b', 2., 42], ['b', 3., 43], ['a', 4., 43]], columns=['name', 'number', 'magic'])

    :py:`DataFrame.groupby` groups independent of the order
    >>> list(df.groupby('name')) # doctest: +NORMALIZE_WHITESPACE
    [('a',
        name  number  magic
      0    a     0.0     42
      1    a     1.0     42
      4    a     4.0     43),
     ('b',
        name  number  magic
      2    b     2.0     42
      3    b     3.0     43)]

    >>> list(consecutive_groupby(df, 'name')) # doctest: +NORMALIZE_WHITESPACE
    [('a',
        name  number  magic
      0    a     0.0     42
      1    a     1.0     42),
     ('b',
        name  number  magic
      2    b     2.0     42
      3    b     3.0     43),
     ('a',
        name  number  magic
      4    a     4.0     43)]

    >>> list(df.groupby(['name', 'magic'])) # doctest: +NORMALIZE_WHITESPACE
    [(('a', np.int64(42)),
        name  number  magic
      0    a     0.0     42
      1    a     1.0     42),
     (('a', np.int64(43)),
        name  number  magic
      4    a     4.0     43),
     (('b', np.int64(42)),
        name  number  magic
      2    b     2.0     42),
     (('b', np.int64(43)),
        name  number  magic
      3    b     3.0     43)]

    >>> list(consecutive_groupby(df, ['name', 'magic'])) # doctest: +NORMALIZE_WHITESPACE
    [(('a', 42),
        name  number  magic
      0    a     0.0     42
      1    a     1.0     42),
     (('b', 42),
        name  number  magic
      2    b     2.0     42),
     (('b', 43),
        name  number  magic
      3    b     3.0     43),
     (('a', 43),
        name  number  magic
      4    a     4.0     43)]

    Returns
    -------
    Iterator over (group_value, group_dataframe). Group value is a scalar if columns is a scalar.
    Otherwise, its a tuple.
    """
    group_vals: pd.DataFrame = df[columns]

    splits = np.not_equal(group_vals.values[1:, ...], group_vals.values[:-1, ...])
    if splits.ndim > 1:
        splits = splits.any(axis=1)
        def get_group_val(i): return tuple(group_vals.values[i])
    else:
        get_group_val = group_vals.values.__getitem__

    split_idx = np.flatnonzero(splits)
    split_idx += 1

    start_idx = 0
    for idx in split_idx:
        group_val = get_group_val(start_idx)
        yield group_val, df.iloc[start_idx:idx, :]
        start_idx = idx

    group_val = get_group_val(start_idx)
    yield group_val, df.iloc[start_idx:, :]



"""
This module contains random functions
"""
import numpy as np
from numpy import ndarray


def random_hermitian(d: int = 2, n: int = 1, std: float = 1,
                     mean: float = 0) -> ndarray:
    r"""
    Generate *n* random Hermitian matrices of dimension *d* drawn from a
    Gaussian ensemble :math:`~\mathcal{N}(\mu, \sigma^2)`.
    """
    H = np.random.randn(n, d, d)*(1 + 1j)*std/2
    H += H.conj().swapaxes(1, 2)
    return H.squeeze() + mean



r"""
This module defines constants, mostly wrapped from scipy. Many constants are in
the top-level namespace. Many more are in the dictionary ``physical_constants``
that can be searched using :meth:`find` and accessed using :meth:`value`.

On top of the scipy constants, the following shorthands are defined:

``e2`` : :math:`e^2`

``pi2`` : :math:`\pi/2`

``pi4`` : :math:`\pi/4`

``two_pi`` : :math:`2\pi`

``four_pi`` : :math:`4\pi`

"""

from typing import Union
import numpy as np
from scipy.constants import (
    Avogadro,
    Boltzmann,
    Btu,
    Btu_IT,
    Btu_th,
    G,
    Julian_year,
    N_A,
    Planck,
    R,
    Rydberg,
    Stefan_Boltzmann,
    Wien,
    acre,
    alpha,
    angstrom,
    arcmin,
    arcminute,
    arcsec,
    arcsecond,
    astronomical_unit,
    atm,
    atmosphere,
    atomic_mass,
    atto,
    au,
    bar,
    barrel,
    bbl,
    blob,
    c,
    calorie,
    calorie_IT,
    calorie_th,
    carat,
    centi,
    convert_temperature,
    day,
    deci,
    degree,
    degree_Fahrenheit,
    deka,
    dyn,
    dyne,
    e,
    eV,
    electron_mass,
    electron_volt,
    elementary_charge,
    epsilon_0,
    erg,
    exa,
    exbi,
    femto,
    fermi,
    find,
    fine_structure,
    fluid_ounce,
    fluid_ounce_US,
    fluid_ounce_imp,
    foot,
    g,
    gallon,
    gallon_US,
    gallon_imp,
    gas_constant,
    gibi,
    giga,
    golden,
    golden_ratio,
    grain,
    gram,
    gravitational_constant,
    h,
    hbar,
    hectare,
    hecto,
    horsepower,
    hour,
    hp,
    inch,
    k,
    kgf,
    kibi,
    kilo,
    kilogram_force,
    kmh,
    knot,
    lambda2nu,
    lb,
    lbf,
    light_year,
    liter,
    litre,
    long_ton,
    m_e,
    m_n,
    m_p,
    m_u,
    mach,
    mebi,
    mega,
    metric_ton,
    micro,
    micron,
    mil,
    mile,
    milli,
    minute,
    mmHg,
    mph,
    mu_0,
    nano,
    nautical_mile,
    neutron_mass,
    nu2lambda,
    ounce,
    oz,
    parsec,
    pebi,
    peta,
    physical_constants,
    pi,
    pico,
    point,
    pound,
    pound_force,
    precision,
    proton_mass,
    psi,
    pt,
    short_ton,
    sigma,
    slinch,
    slug,
    speed_of_light,
    speed_of_sound,
    stone,
    survey_foot,
    survey_mile,
    tebi,
    tera,
    ton_TNT,
    torr,
    troy_ounce,
    troy_pound,
    u,
    unit,
    value,
    week,
    yard,
    year,
    yobi,
    yotta,
    zebi,
    zepto,
    zero_Celsius,
    zetta
)

e2 = e**2
pi2 = pi/2
pi4 = pi/4
two_pi = pi*2
four_pi = pi*4


def convert_attenuation(val: float, old_scale: str, new_scale: str) -> float:
    """Convert between different units of attenuation.

    Parameters
    ----------
    val: float
        Value to be converted.
    old_scale: str
        dB, field, or power.
    new_scale: str
        dB, field, or power.

    Returns
    -------
    The converted values.

    """

    if old_scale.lower() == 'db':
        atten = np.asanyarray(val)
    elif old_scale.lower() in ['field', 'f']:
        atten = 20*np.log10(np.asanyarray(val))
    elif old_scale.lower() in ['power', 'p']:
        atten = 10*np.log10(np.asanyarray(val))

    if new_scale.lower() == 'db':
        res = atten
    elif new_scale.lower() in ['field', 'f']:
        res = 10**(atten / 20)
    elif new_scale.lower() in ['power', 'p']:
        res = 10**(atten / 10)

    if res.ndim == 0:
        return float(res)

    return res


def convert_power(val: float, old_scale: str, new_scale: str) -> float:
    """Convert between different units of power.

    Parameters
    ----------
    val: float
        Value to be converted.
    old_scale: str
        dBm, watt, Vpp, or Vrms.
    new_scale: str
        dBm, watt, Vpp, or Vrms.

    Returns
    -------
    The converted values.

    """

    if old_scale.lower() == 'dbm':
        power = np.asanyarray(val)
    elif old_scale.lower() in ['watt', 'w']:
        power = 10 + 20*np.log10(np.sqrt(np.asanyarray(val)*100));
    elif old_scale.lower() == 'vpp':
        power = 10 + 20*np.log10(np.asanyarray(val)/2);
    elif old_scale.lower() == 'vrms':
        power = 10 + 20*np.log10(np.asanyarray(val)*np.sqrt(2))

    if new_scale.lower() == 'dbm':
        res = power
    elif new_scale.lower() in ['watt', 'w']:
        res = 10**((power - 10) / 20)**2 / 100;
    elif new_scale.lower() == 'vpp':
        res = 2 * 10**((power - 10) / 20)
    elif new_scale.lower() == 'vrms':
        res = 10**((power - 10) / 20) / np.sqrt(2);

    if res.ndim == 0:
        return float(res)

    return res


def lambda2eV(lambda_: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """Convert wavelength to eV.

    Parameters
    ----------
    lambda_: Union[float, np.ndarray]
        Wavelength in meters.

    Returns
    -------
    eV: Union[float, np.ndarray]
        Energy in eV.

    """
    return lambda2nu(lambda_)*h/e


def eV2lambda(eV: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """Convert eV to wavelength.

    Parameters
    ----------
    eV: Union[float, np.ndarray]
        Energy in eV.

    Returns
    -------
    lambda: Union[float, np.ndarray]
        Wavelength in meters.

    """
    return nu2lambda(e/h*eV)



"""This module contains signal processing functions that work on data in
frequency space. For consistency, functions in this module should have
the data to be processed as first and the frequencies at which the data
is sampled as their second argument, while always returning a tuple of
(possibly processed) frequencies and processed data, i.e.::

    def fun(data, f, *args, **kwargs) -> processed_data, processed_f:
        ...

"""
import inspect
from typing import Literal, Tuple, TypeVar, Optional

import numpy as np
from scipy import integrate

from qutil import math
from qutil.functools import wraps
from qutil.typecheck import check_literals

try:
    import numba
except ImportError:
    numba = None

_S = TypeVar('_S')
_T = TypeVar('_T')


def _standardize(function):
    """Adds variadic kwargs and f arg and return param."""
    try:
        parameters = inspect.signature(function).parameters
        assert 'f' not in parameters, \
            'Only use this decorator for functions without parameter named f'
        assert not any(p.kind is inspect.Parameter.VAR_KEYWORD for p in parameters.values()), \
            'Only use this decorator for functions without variadic keyword arguments.'
    except ValueError:
        # ufunc, https://numpy.org/doc/stable/reference/ufuncs.html#ufuncs-kwargs
        parameters = {'out', 'where', 'axes', 'axis', 'keepdims', 'casting', 'order', 'dtype',
                      'subok', 'signature', 'extobj'}

    @wraps(function)
    def wrapper(x, f, *args, **kwargs):
        # Filter kwargs that function actually accepts
        kwargs = {k: v for k, v in kwargs.items() if k in parameters}
        return function(x, *args, **kwargs), f

    return wrapper


def Id(x: _S, f: _T, *_, **__) -> Tuple[_S, _T]:
    """The identity mapping."""
    return x, f


def derivative(x, f, deriv_order: int = 0, **_) -> Tuple[np.ndarray, np.ndarray]:
    """Perform (anti-)derivatives.

    .. note::
        For negative antiderivatives, the zero-frequency component is
        set to zero (due to zero-division).

    Parameters
    ----------
    x : array_like
        The data to be filtered.
    f : array_like
        Frequencies corresponding to the last axis of `x`.
    deriv_order : int
        The order of the derivative. If negative, the antiderivative is
        computed (indefinite integral). Default: 0.
    """
    f = np.asanyarray(f)
    with np.errstate(invalid='ignore', divide='ignore'):
        xp = np.asanyarray(x)*(2*np.pi*f)**deriv_order
    if deriv_order < 0:
        xp[..., (f == 0).nonzero()] = 0
    return xp, f


def rms(x, f, /, out=None, *, axis: Optional[int] = None, where=True, dtype=None, keepdims=False,
        **_) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the RMS (root-mean-square).

    See :class:`numpy.ufunc` and the `NumPy reference`_ for
    documentation of the arguments.

    .. _NumPy reference: https://numpy.org/doc/stable/reference/ufuncs.html

    Examples
    --------
    >>> t = np.linspace(0, 1, 1001)
    >>> x = 2*np.sqrt(2)*np.sin(2*np.pi*10*t)
    >>> xf = np.fft.fft(x)  # nb rfft would need to be scaled by factor √2
    >>> r, _ = rms(xf, ...)  # f not actually needed
    >>> print(r)  # doctest: +ELLIPSIS
    1.9990007493755...
    >>> np.allclose(r, np.sqrt(x.mean()**2 + x.var()))
    True
    """
    x = np.asanyarray(x)
    N = np.take(x.shape, axis or range(x.ndim)).prod()

    result = math.abs2(x, where=where, dtype=dtype)
    result = np.sum(result, axis=axis, dtype=dtype, out=out, keepdims=keepdims, where=where)
    result = np.sqrt(result, out=out, dtype=dtype)
    result /= N
    return result, f


if numba is not None:
    # nb.guvectorize generates gufuncs with all the kwargs, so only work and result array required.
    def _rms(x, res):
        res[0] = 0
        if np.iscomplexobj(x):
            for i in range(x.shape[0]):
                xx = x[i]
                real = xx.real
                imag = xx.imag
                res += real * real + imag * imag
        else:
            for i in range(x.shape[0]):
                real = x[i].real
                res += real * real
        res[0] = np.sqrt(res[0]) / x.shape[0]

    _rms.__doc__ = rms.__doc__
    # Expose both the generated ufunc and the wrapped version that complies with the signature
    # convention.
    rms_ufunc = numba.guvectorize([(numba.float32[:], numba.float32[:]),
                                   (numba.float64[:], numba.float64[:]),
                                   (numba.complex64[:], numba.float32[:]),
                                   (numba.complex128[:], numba.float64[:])],
                                  '(n)->()',
                                  target='parallel',
                                  cache=True)(_rms)
    try:
        rms = _standardize(rms_ufunc)
    except AssertionError:
        # Could not parse rms_ufunc signature, leave rms alone
        pass


def brickwall_filter(x, f, f_min: float = 0, f_max: float = np.inf, **_) -> Tuple[np.ndarray,
                                                                                  np.ndarray]:
    """Apply a brick wall filter to the data.

    Parameters
    ----------
    x : array_like
        The data to be filtered.
    f : array_like
        Frequencies corresponding to the last axis of `x`.
    f_min, f_max : float
        The locations of the brick walls for (low-, band-, high-)pass
        filtering.
    """
    f = np.asanyarray(f)
    xp = np.copy(x)
    xp[..., (f < f_min) | (f > f_max)] = 0
    return xp, f


@check_literals
def octave_band_rms(x, f, base: Literal[2, 10] = 10, fraction: int = 1,
                    return_band_edges: bool = False, **_) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the rms over octave fractions [1]_, [2]_.

    Parameters
    ----------
    x : ndarray, shape (..., n_freq)
        The amplitude spectral density to compute the rms from.
        (I.e., the frequency-domain data).
    f : ndarray, shape (n_freq,)
        The frequencies corresponding to the data x.
    base : 2 or 10
        The logarithm base for the octaves.
    fraction : int
        The denominator of the fraction of an octave band to use for
        the calculation.
    return_band_edges : bool
        Return the edges instead of the midband frequencies.

    Returns
    -------
    octave_band_rms : ndarray
        The rms values.
    bandedge_frequencies : ndarray
        The frequencies of the octave band edges
        (if return_band_edges is true).
    midband_frequencies : ndarray
        The midband frequencies of the octave bands.
        (if return_band_edges is false).

    References
    ----------
    .. [1] C. G. Gordon, “Generic vibration criteria for vibration-
       sensitive equipment”, in Optomechanical Engineering and
       Vibration Control, Vol. 3786 (Sept. 28, 1999), pp. 22–33.
       https://www.spiedigitallibrary.org/conference-proceedings-of-spie/3786/0000/Generic-vibration-criteria-for-vibration-sensitive-equipment/10.1117/12.363802.short
    .. [2] ANSI/ASA S1.11-2004 (R2009). Octave-Band and Fractional-
       Octave-Band Analog and Digital Filters.
       https://webstore.ansi.org/standards/asa/ansiasas1112004r2009
    """
    mask = f > 0
    f = f[mask]
    x = math.abs2(x[..., mask])
    df = f[1] - f[0]

    # Computed according to ANSI S1.11-2004
    reference_frequency = 1000
    frequency_ratio = 10**0.3 if base == 10 else 2
    # Compute the band index x from the frequencies given
    # and then calculate back the midband frequencies
    if fraction % 2:
        bmin = fraction*np.log(f.min()/reference_frequency)/np.log(frequency_ratio) + 30
        bmax = fraction*np.log(f.max()/reference_frequency)/np.log(frequency_ratio) + 30
        band_index = np.arange(np.ceil(bmin), np.ceil(bmax))
        midband_frequencies = reference_frequency*frequency_ratio**((band_index - 30)/fraction)
    else:
        bmin = (2*fraction*np.log(f.min()/reference_frequency)/np.log(frequency_ratio) + 59)/2
        bmax = (2*fraction*np.log(f.max()/reference_frequency)/np.log(frequency_ratio) + 59)/2
        band_index = np.arange(np.ceil(bmin), np.ceil(bmax))
        midband_frequencies = reference_frequency*frequency_ratio**((2*band_index - 59)/2/fraction)

    bandedge_frequencies = (midband_frequencies[:, None]
                            * frequency_ratio**(np.array([-1, 1])/2/fraction))
    bandwidths = np.diff(bandedge_frequencies)[:, 0]

    # drop bands for which the frequency resolution is too low to integrate
    mask = 2*df < bandwidths
    midband_frequencies = midband_frequencies[mask]
    bandedge_frequencies = bandedge_frequencies[mask, :]

    mean_square = np.empty(x.shape[:-1] + midband_frequencies.shape)
    for i, (f_lower, f_upper) in enumerate(bandedge_frequencies):
        mask = (f_lower <= f) & (f <= f_upper)
        mean_square[..., i] = integrate.trapezoid(x[..., mask], f[..., mask])

    if return_band_edges:
        return np.sqrt(mean_square), bandedge_frequencies
    else:
        return np.sqrt(mean_square), midband_frequencies
